import logging
import operator
from collections.abc import Callable
from copy import deepcopy
from enum import StrEnum
from typing import Literal, overload
from warnings import warn

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix

from .exception import TimestampAttributeMissingError


logger = logging.getLogger(__name__)


class ItemUserBasedEnum(StrEnum):
    """Enum class for item and user based properties.

    Enum class to indicate if the function or logic is based on item or user.
    """

    ITEM = "item"
    """Property based on item"""
    USER = "user"
    """Property based on user"""

    @classmethod
    def has_value(cls, value: str) -> bool:
        """Check valid value for ItemUserBasedEnum

        :param value: String value input
        :type value: str
        """
        return value in ItemUserBasedEnum


class InteractionMatrix:
    """Matrix of interaction data between users and items.

    It provides a number of properties and methods for easy manipulation of this interaction data.

    .. attention::

        - The InteractionMatrix does not assume binary user-item pairs.
          If a user interacts with an item more than once, there will be two
          entries for this user-item pair.

        - We assume that the user and item IDs are integers starting from 0. IDs
          that are indicated by "-1" are reserved to label the user or item to
          be predicted. This assumption is crucial as it will be used during the
          split scheme and evaluation of the RS since it will affect the 2D shape
          of the CSR matrix

    :param df: Dataframe containing user-item interactions. Must contain at least
        item ids and user ids.
    :type df: pd.DataFrame
    :param item_ix: Item ids column name.
    :type item_ix: str
    :param user_ix: User ids column name.
    :type user_ix: str
    :param timestamp_ix: Interaction timestamps column name.
    :type timestamp_ix: str
    :param shape: The desired shape of the matrix, i.e. the number of users and items.
        If no shape is specified, the number of users will be equal to the
        maximum user id plus one, the number of items to the maximum item
        id plus one.
    :type shape: tuple[int, int], optional
    :param skip_df_processing: Skip processing of the dataframe. This is useful
        when the dataframe is already processed and the columns are already
        renamed.
    :type skip_df_processing: bool, optional
    """

    ITEM_IX = "iid"
    USER_IX = "uid"
    TIMESTAMP_IX = "ts"
    INTERACTION_IX = "interactionid"
    MASKED_LABEL = -1

    def __init__(
        self,
        df: pd.DataFrame,
        item_ix: str,
        user_ix: str,
        timestamp_ix: str,
        shape: None | tuple[int, int] = None,
        skip_df_processing: bool = False,
    ) -> None:
        self.user_item_shape: tuple[int, int]
        """The shape of the interaction matrix, i.e. `|user| x |item|`."""
        if shape:
            self.user_item_shape = shape

        if skip_df_processing:
            self._df = df
            return

        col_mapper = {
            item_ix: InteractionMatrix.ITEM_IX,
            user_ix: InteractionMatrix.USER_IX,
            timestamp_ix: InteractionMatrix.TIMESTAMP_IX,
        }
        df = df.rename(columns=col_mapper)
        required_columns = [
            InteractionMatrix.USER_IX,
            InteractionMatrix.ITEM_IX,
            InteractionMatrix.TIMESTAMP_IX,
        ]
        extra_columns = [col for col in df.columns if col not in required_columns]
        df = df[required_columns + extra_columns].copy()
        # TODO refactor this statement below
        df = df.reset_index(drop=True).reset_index().rename(columns={"index": InteractionMatrix.INTERACTION_IX})

        self._df = df

    def copy(self) -> "InteractionMatrix":
        """Create a deep copy of this InteractionMatrix.

        :return: Deep copy of this InteractionMatrix.
        :rtype: InteractionMatrix
        """
        return deepcopy(self)

    def copy_df(self, reset_index: bool = False) -> "pd.DataFrame":
        """Create a deep copy of the dataframe.

        :return: Deep copy of dataframe.
        :rtype: pd.DataFrame
        """
        if reset_index:
            return deepcopy(self._df.reset_index(drop=True))
        return deepcopy(self._df)

    def concat(self, im: "InteractionMatrix | pd.DataFrame") -> "InteractionMatrix":
        """Concatenate this InteractionMatrix with another.

        .. note::
            This is a inplace operation. and will modify the current object.

        :param im: InteractionMatrix to concat with.
        :type im: Union[InteractionMatrix, pd.DataFrame]
        :return: InteractionMatrix with the interactions from both matrices.
        :rtype: InteractionMatrix
        """
        if isinstance(im, pd.DataFrame):
            self._df = pd.concat([self._df, im])
        else:
            self._df = pd.concat([self._df, im._df])

        return self

    # TODO this should be shifted to prediction matrix
    def union(self, im: "InteractionMatrix") -> "InteractionMatrix":
        """Combine events from this InteractionMatrix with another.

        :param im: InteractionMatrix to union with.
        :type im: InteractionMatrix
        :return: Union of interactions in this InteractionMatrix and the other.
        :rtype: InteractionMatrix
        """
        return self + im

    def difference(self, im: "InteractionMatrix") -> "InteractionMatrix":
        """Difference between this InteractionMatrix and another.

        :param im: InteractionMatrix to subtract from this.
        :type im: InteractionMatrix
        :return: Difference between this InteractionMatrix and the other.
        :rtype: InteractionMatrix
        """
        return self - im

    @property
    def values(self) -> csr_matrix:
        """All user-item interactions as a sparse matrix of size (|`global_users`|, |`global_items`|).

        Each entry is the number of interactions between that user and item.
        If there are no interactions between a user and item, the entry is 0.

        :return: Interactions between users and items as a csr_matrix.
        :rtype: csr_matrix
        """
        # TODO issue with -1 labeling in the interaction matrix should i create prediction matrix
        if not hasattr(self, "user_item_shape"):
            raise AttributeError("InteractionMatrix has no `user_item_shape` attribute. Please call mask_shape() first.")

        values = np.ones(self._df.shape[0])
        indices = self._df[[InteractionMatrix.USER_IX, InteractionMatrix.ITEM_IX]].values
        indices = (indices[:, 0], indices[:, 1])

        matrix = csr_matrix((values, indices), shape=self.user_item_shape, dtype=np.int32)
        return matrix

    @property
    def indices(self) -> tuple[list[int], list[int]]:
        """Returns a tuple of lists of user IDs and item IDs corresponding to interactions.

        :return: tuple of lists of user IDs and item IDs that correspond to at least one interaction.
        :rtype: tuple[list[int], list[int]]
        """
        return self.values.nonzero()

    def nonzero(self) -> tuple[list[int], list[int]]:
        return self.values.nonzero()

    @overload
    def users_in(self, U: set[int]) -> "InteractionMatrix": ...
    @overload
    def users_in(self, U: set[int], inplace: Literal[False]) -> "InteractionMatrix": ...
    @overload
    def users_in(self, U: set[int], inplace: Literal[True]) -> None: ...
    def users_in(self, U: set[int], inplace: bool = False) -> "None | InteractionMatrix":
        """Keep only interactions by one of the specified users.

        :param U: A set or list of users to select the interactions from.
        :type U: Union[set[int]
        :param inplace: Apply the selection in place or not, defaults to False
        :type inplace: bool, optional
        :return: None if `inplace`, otherwise returns a new InteractionMatrix object
        :rtype: Union[InteractionMatrix, None]
        """
        logger.debug("Performing users_in comparison")

        mask = self._df[InteractionMatrix.USER_IX].isin(U)

        return self._apply_mask(mask, inplace=inplace)

    @overload
    def _apply_mask(self, mask: pd.Series) -> "InteractionMatrix": ...
    @overload
    def _apply_mask(self, mask: pd.Series, inplace: Literal[True]) -> None: ...
    @overload
    def _apply_mask(self, mask: pd.Series, inplace: Literal[False]) -> "InteractionMatrix": ...
    def _apply_mask(self, mask: pd.Series, inplace: bool = False) -> "None | InteractionMatrix":
        interaction_m = self if inplace else self.copy()
        interaction_m._df = interaction_m._df[mask]
        return None if inplace else interaction_m

    def _timestamps_cmp(self, op: Callable, timestamp: float, inplace: bool = False) -> "None | InteractionMatrix":
        """Filter interactions based on timestamp.
        Keep only interactions for which op(t, timestamp) is True.

        :param op: Comparison operator.
        :type op: Callable
        :param timestamp: Timestamp to compare against in seconds from epoch.
        :type timestamp: float
        :param inplace: Modify the data matrix in place. If False, returns a new object.
        :type inplace: bool, optional
        """
        logger.debug(f"Performing {op.__name__}(t, {timestamp})")

        mask = op(self._df[InteractionMatrix.TIMESTAMP_IX], timestamp)
        return self._apply_mask(mask, inplace=inplace)

    @overload
    def timestamps_gt(self, timestamp: float) -> "InteractionMatrix": ...
    @overload
    def timestamps_gt(self, timestamp: float, inplace: Literal[True]) -> None: ...
    def timestamps_gt(self, timestamp: float, inplace: bool = False) -> "None | InteractionMatrix":
        """Select interactions after a given timestamp.

        :param timestamp: The timestamp with which
            the interactions timestamp is compared.
        :type timestamp: float
        :param inplace: Apply the selection in place if True, defaults to False
        :type inplace: bool, optional
        :return: None if `inplace`, otherwise returns a new InteractionMatrix object
        :rtype: Union[InteractionMatrix, None]
        """
        return self._timestamps_cmp(operator.gt, timestamp, inplace)

    @overload
    def timestamps_gte(self, timestamp: float) -> "InteractionMatrix": ...
    @overload
    def timestamps_gte(self, timestamp: float, inplace: Literal[True]) -> None: ...
    def timestamps_gte(self, timestamp: float, inplace: bool = False) -> "None | InteractionMatrix":
        """Select interactions after and including a given timestamp.

        :param timestamp: The timestamp with which
            the interactions timestamp is compared.
        :type timestamp: float
        :param inplace: Apply the selection in place if True, defaults to False
        :type inplace: bool, optional
        :return: None if `inplace`, otherwise returns a new InteractionMatrix object
        :rtype: Union[InteractionMatrix, None]
        """
        return self._timestamps_cmp(operator.ge, timestamp, inplace)

    @overload
    def timestamps_lt(self, timestamp: float) -> "InteractionMatrix": ...
    @overload
    def timestamps_lt(self, timestamp: float, inplace: Literal[True]) -> None: ...
    def timestamps_lt(self, timestamp: float, inplace: bool = False) -> "None | InteractionMatrix":
        """Select interactions up to a given timestamp.

        :param timestamp: The timestamp with which
            the interactions timestamp is compared.
        :type timestamp: float
        :param inplace: Apply the selection in place if True, defaults to False
        :type inplace: bool, optional
        :return: None if `inplace`, otherwise returns a new InteractionMatrix object
        :rtype: Union[InteractionMatrix, None]
        """
        return self._timestamps_cmp(operator.lt, timestamp, inplace)

    @overload
    def timestamps_lte(self, timestamp: float) -> "InteractionMatrix": ...
    @overload
    def timestamps_lte(self, timestamp: float, inplace: Literal[True]) -> None: ...
    def timestamps_lte(self, timestamp: float, inplace: bool = False) -> "None | InteractionMatrix":
        """Select interactions up to and including a given timestamp.

        :param timestamp: The timestamp with which
            the interactions timestamp is compared.
        :type timestamp: float
        :param inplace: Apply the selection in place if True, defaults to False
        :type inplace: bool, optional
        :return: None if `inplace`, otherwise returns a new InteractionMatrix object
        :rtype: Union[InteractionMatrix, None]
        """
        return self._timestamps_cmp(operator.le, timestamp, inplace)

    def __add__(self, im: "InteractionMatrix") -> "InteractionMatrix":
        """Combine events from this InteractionMatrix with another.

        :param im: InteractionMatrix to union with.
        :type im: InteractionMatrix
        :return: Union of interactions in this InteractionMatrix and the other.
        :rtype: InteractionMatrix
        """
        df = pd.concat([self._df, im._df], copy=False)

        shape = None
        if hasattr(self, "user_item_shape") and hasattr(im, "user_item_shape"):
            shape = (max(self.user_item_shape[0], im.user_item_shape[0]), max(self.user_item_shape[1], im.user_item_shape[1]))
            self.user_item_shape = shape

        return InteractionMatrix(
            df,
            InteractionMatrix.ITEM_IX,
            InteractionMatrix.USER_IX,
            InteractionMatrix.TIMESTAMP_IX,
            shape,
            True,
        )

    def __sub__(self, im: "InteractionMatrix") -> "InteractionMatrix":
        full_data = pd.MultiIndex.from_frame(self._df)
        data_part_2 = pd.MultiIndex.from_frame(im._df)
        data_part_1 = full_data.difference(data_part_2).to_frame().reset_index(drop=True)

        shape = None
        if hasattr(self, "user_item_shape") and hasattr(im, "user_item_shape"):
            shape = (max(self.user_item_shape[0], im.user_item_shape[0]), max(self.user_item_shape[1], im.user_item_shape[1]))

        return InteractionMatrix(
            data_part_1,
            InteractionMatrix.ITEM_IX,
            InteractionMatrix.USER_IX,
            InteractionMatrix.TIMESTAMP_IX,
            shape,
            True,
        )

    def __repr__(self) -> str:
        return repr(self._df)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, InteractionMatrix):
            logger.debug(f"Comparing {type(value)} with InteractionMatrix is not supported")
            return False
        return self._df.equals(value._df)

    def __len__(self) -> int:
        """Return the number of interactions in the matrix.

        This is distinct from the shape of the matrix, which is the number of
        users and items that has been released to the model. The length of the
        matrix is the number of interactions present in the matrix resulting
        from filter operations.
        """
        return len(self._df)

    @overload
    def items_in(self, id_set: set[int]) -> "InteractionMatrix": ...
    @overload
    def items_in(self, id_set: set[int], inplace: Literal[False]) -> "InteractionMatrix": ...
    @overload
    def items_in(self, id_set: set[int], inplace: Literal[True]) -> None: ...
    def items_in(self, id_set: set[int], inplace=False) -> "None | InteractionMatrix":
        """Keep only interactions with the specified items.

        :param id_set: A set or list of items to select the interactions.
        :type id_set: set[int]
        :param inplace: Apply the selection in place or not, defaults to False
        :type inplace: bool, optional
        :return: None if `inplace`, otherwise returns a new InteractionMatrix object
        :rtype: Union[InteractionMatrix, None]
        """
        logger.debug("Performing items_in comparison")

        mask = self._df[InteractionMatrix.ITEM_IX].isin(id_set)

        return self._apply_mask(mask, inplace=inplace)

    @overload
    def items_not_in(self, id_set: set[int]) -> "InteractionMatrix": ...
    @overload
    def items_not_in(self, id_set: set[int], inplace: Literal[False]) -> "InteractionMatrix": ...
    @overload
    def items_not_in(self, id_set: set[int], inplace: Literal[True]) -> None: ...
    def items_not_in(self, id_set: set[int], inplace: bool=False) -> "None | InteractionMatrix":
        """Keep only interactions not with the specified items.

        :param id_set: A set or list of items to exclude from the interactions.
        :type id_set: set[int]
        :param inplace: Apply the selection in place or not, defaults to False
        :type inplace: bool, optional
        :return: None if `inplace`, otherwise returns a new InteractionMatrix object
        :rtype: Union[InteractionMatrix, None]
        """
        logger.debug("Performing items_not_in comparison")

        mask = ~self._df[InteractionMatrix.ITEM_IX].isin(id_set)

        return self._apply_mask(mask, inplace=inplace)

    def users_not_in(self, U: set[int], inplace=False) -> "None | InteractionMatrix":
        """Keep only interactions not by the specified users.

        :param U: A set or list of users to exclude from the interactions.
        :type U: set[int]
        :param inplace: Apply the selection in place or not, defaults to False
        :type inplace: bool, optional
        :return: None if `inplace`, otherwise returns a new InteractionMatrix object
        :rtype: Union[InteractionMatrix, None]
        """
        logger.debug("Performing users_not_in comparison")

        mask = ~self._df[InteractionMatrix.USER_IX].isin(U)

        return self._apply_mask(mask, inplace=inplace)

    def interactions_in(self, interaction_ids: list[int], inplace: bool = False) -> "None | InteractionMatrix":
        """Select the interactions by their interaction ids

        :param interaction_ids: A list of interaction ids
        :type interaction_ids: list[int]
        :param inplace: Apply the selection in place,
            or return a new InteractionMatrix object, defaults to False
        :type inplace: bool, optional
        :return: None if inplace, otherwise new InteractionMatrix
            object with the selected interactions
        :rtype: Union[None, InteractionMatrix]
        """
        logger.debug("Performing interactions_in comparison")

        mask = self._df[InteractionMatrix.INTERACTION_IX].isin(interaction_ids)

        unknown_interaction_ids = set(interaction_ids).difference(self._df[InteractionMatrix.INTERACTION_IX].unique())

        if unknown_interaction_ids:
            warn(f"IDs {unknown_interaction_ids} not present in data")
        if not interaction_ids:
            warn("No interaction IDs given, returning empty InteractionMatrix.")

        return self._apply_mask(mask, inplace=inplace)

    def _get_last_n_interactions(
        self,
        by: ItemUserBasedEnum,
        n_seq_data: int,
        t_upper: None | int = None,
        id_in: None | set[int] = None,
        inplace=False,
    ) -> "InteractionMatrix":
        if not self.has_timestamps:
            raise TimestampAttributeMissingError()
        if t_upper is None:
            t_upper = self.max_timestamp + 1  # to include the last timestamp

        interaction_m = self if inplace else self.copy()

        mask = interaction_m._df[InteractionMatrix.TIMESTAMP_IX] < t_upper
        if id_in and by == ItemUserBasedEnum.USER:
            mask = mask & interaction_m._df[InteractionMatrix.USER_IX].isin(id_in)
        elif id_in and by == ItemUserBasedEnum.ITEM:
            mask = mask & interaction_m._df[InteractionMatrix.ITEM_IX].isin(id_in)

        if by == ItemUserBasedEnum.USER:
            c_df = interaction_m._df[mask].groupby(InteractionMatrix.USER_IX).tail(n_seq_data)
        else:
            c_df = interaction_m._df[mask].groupby(InteractionMatrix.ITEM_IX).tail(n_seq_data)
        interaction_m._df = c_df

        return interaction_m

    def _get_first_n_interactions(
        self, by: ItemUserBasedEnum, n_seq_data: int, t_lower: None | int = None, inplace=False
    ) -> "InteractionMatrix":
        if not self.has_timestamps:
            raise TimestampAttributeMissingError()
        if t_lower is None:
            t_lower = self.min_timestamp

        interaction_m = self if inplace else self.copy()

        mask = interaction_m._df[InteractionMatrix.TIMESTAMP_IX] >= t_lower
        if by == ItemUserBasedEnum.USER:
            c_df = interaction_m._df[mask].groupby(InteractionMatrix.USER_IX).head(n_seq_data)
        else:
            c_df = interaction_m._df[mask].groupby(InteractionMatrix.ITEM_IX).head(n_seq_data)
        interaction_m._df = c_df
        return interaction_m

    def get_users_n_last_interaction(
        self,
        n_seq_data: int = 1,
        t_upper: None | int = None,
        user_in: None | set[int] = None,
        inplace: bool = False,
    ) -> "InteractionMatrix":
        """Select the last n interactions for each user.

        :param n_seq_data: Number of interactions to select, defaults to 1
        :type n_seq_data: int, optional
        :param t_upper: Seconds past t. Upper limit for the timestamp
            of the interactions to select, defaults to None
        :type t_upper: None | int, optional
        :param user_in: set of user IDs to select the interactions from,
            defaults to None
        :type user_in: None | set[int], optional
        :param inplace: If operation is inplace, defaults to False
        :type inplace: bool, optional
        :return: Resulting interaction matrix
        :rtype: InteractionMatrix
        """
        logger.debug("Performing get_user_n_last_interaction comparison")
        return self._get_last_n_interactions(ItemUserBasedEnum.USER, n_seq_data, t_upper, user_in, inplace)

    def get_items_n_last_interaction(
        self,
        n_seq_data: int = 1,
        t_upper: None | int = None,
        item_in: None | set[int] = None,
        inplace: bool = False,
    ) -> "InteractionMatrix":
        """Select the last n interactions for each item.

        :param n_seq_data: Number of interactions to select, defaults to 1
        :type n_seq_data: int, optional
        :param t_upper: Seconds past t. Upper limit for the timestamp
            of the interactions to select, defaults to None
        :type t_upper: None | int, optional
        :param item_in: set of item IDs to select the interactions from,
            defaults to None
        :type item_in: None | set[int], optional
        :param inplace: If operation is inplace, defaults to False
        :type inplace: bool, optional
        :return: Resulting interaction matrix
        :rtype: InteractionMatrix
        """
        logger.debug("Performing get_item_n_last_interaction comparison")
        return self._get_last_n_interactions(ItemUserBasedEnum.ITEM, n_seq_data, t_upper, item_in, inplace)

    def get_users_n_first_interaction(
        self, n_seq_data: int = 1, t_lower: None | int = None, inplace=False
    ) -> "InteractionMatrix":
        """Select the first n interactions for each user.

        :param n_seq_data: Number of interactions to select, defaults to 1
        :type n_seq_data: int, optional
        :param t_lower: Seconds past t. Lower limit for the timestamp
            of the interactions to select, defaults to None
        :type t_lower: None | int, optional
        :param inplace: If operation is inplace, defaults to False
        :type inplace: bool, optional
        :return: Resulting interaction matrix
        :rtype: InteractionMatrix
        """
        return self._get_first_n_interactions(ItemUserBasedEnum.USER, n_seq_data, t_lower, inplace)

    def get_items_n_first_interaction(
        self, n_seq_data: int = 1, t_lower: None | int = None, inplace=False
    ) -> "InteractionMatrix":
        """Select the first n interactions for each item.

        :param n_seq_data: Number of interactions to select, defaults to 1
        :type n_seq_data: int, optional
        :param t_lower: Seconds past t. Lower limit for the timestamp
            of the interactions to select, defaults to None
        :type t_lower: None | int, optional
        :param inplace: If operation is inplace, defaults to False
        :type inplace: bool, optional
        :return: Resulting interaction matrix
        :rtype: InteractionMatrix
        """
        return self._get_first_n_interactions(ItemUserBasedEnum.ITEM, n_seq_data, t_lower, inplace)

    @property
    def item_interaction_sequence_matrix(self) -> csr_matrix:
        """Converts the interaction data into an item interaction sequence matrix.

        Dataframe values are converted into such that the row sequence is maintained and
        the item that interacted with will have the column at item_id marked with 1.
        """
        values = np.ones(self._df.shape[0])
        indices = (np.arange(self._df.shape[0]), self._df[InteractionMatrix.ITEM_IX].to_numpy())
        shape = (self._df.shape[0], self.user_item_shape[1])

        sparse_matrix = csr_matrix((values, indices), shape=shape, dtype=values.dtype)
        return sparse_matrix

    @property
    def user_id_sequence_array(self) -> np.ndarray:
        """Array of user IDs in the order of interactions.

        :return: Numpy array of user IDs.
        :rtype: np.ndarray
        """
        return self._df[InteractionMatrix.USER_IX].to_numpy()

    def get_prediction_data(self) -> "InteractionMatrix":
        """Get the data to be predicted.

        :return: InteractionMatrix with only the data to be predicted.
        :rtype: InteractionMatrix
        """
        return self.items_in({-1})

    def get_interaction_data(self) -> "InteractionMatrix":
        """Get the data that is not denoted by "-1".

        User and item IDs that are not denoted by "-1" are the ones that are
        known to the model.

        :return: InteractionMatrix with only the known data.
        :rtype: InteractionMatrix
        """
        mask = (self._df[InteractionMatrix.USER_IX] != -1) & (self._df[InteractionMatrix.ITEM_IX] != -1)
        return self._apply_mask(mask)

    @property
    def user_ids(self) -> set[int]:
        """The set of all user ID in matrix"""
        return set(self._df[InteractionMatrix.USER_IX].dropna().unique())

    @property
    def item_ids(self) -> set[int]:
        """The set of all item ID in matrix"""
        return set(self._df[InteractionMatrix.ITEM_IX].dropna().unique())

    @property
    def num_interactions(self) -> int:
        """The total number of interactions.

        :return: Total interaction count.
        :rtype: int
        """
        return len(self._df)

    @property
    def has_timestamps(self) -> bool:
        """Boolean indicating whether instance has timestamp information.

        :return: True if timestamps information is available, False otherwise.
        :rtype: bool
        """
        return self.TIMESTAMP_IX in self._df

    @property
    def min_timestamp(self) -> int:
        """The earliest timestamp in the interaction

        :return: The earliest timestamp.
        :rtype: int
        """
        if not self.has_timestamps:
            raise TimestampAttributeMissingError()
        return self._df[self.TIMESTAMP_IX].min()

    @property
    def max_timestamp(self) -> int:
        """The latest timestamp in the interaction

        :return: The latest timestamp.
        :rtype: int
        """
        if not self.has_timestamps:
            raise TimestampAttributeMissingError()
        return self._df[self.TIMESTAMP_IX].max()

    @property
    def max_global_user_id(self) -> int:
        """The highest known global user ID in the interaction matrix.

        The difference between `max_global_user_id` and `max_user_id` is that
        `max_global_user_id` considers all user IDs present in the dataframe,
        including users that are only encountered during prediction time.
        """
        return max(int(self._df[InteractionMatrix.USER_IX].max()) + 1, self.user_item_shape[0])

    @property
    def max_global_item_id(self) -> int:
        return max(int(self._df[InteractionMatrix.ITEM_IX].max()) + 1, self.user_item_shape[1])

    @property
    def max_known_user_id(self) -> int:
        """The highest known user ID in the interaction matrix."""
        max_val = self._df[(self._df != -1).all(axis=1)][InteractionMatrix.USER_IX].max()
        if pd.isna(max_val):
            return self.user_item_shape[0]
        return min(int(max_val) + 1, self.user_item_shape[0])

    @property
    def max_known_item_id(self) -> int:
        """The highest known user ID in the interaction matrix."""
        max_val = self._df[(self._df != -1).all(axis=1)][InteractionMatrix.ITEM_IX].max()
        if pd.isna(max_val):
            return self.user_item_shape[1]
        return min(int(max_val) + 1, self.user_item_shape[1])

    @property
    def max_user_id(self) -> int:
        """The highest known user ID in the interaction matrix.

        :return: The highest user ID.
        :rtype: int
        """
        max_val = self._df[self._df != -1][InteractionMatrix.USER_IX].max()
        if np.isnan(max_val):
            return -1
        return max_val

    @property
    def max_item_id(self) -> int:
        """The highest known item ID in the interaction matrix.

        In the case of an empty matrix, the highest item ID is -1. This is
        consistent with the the definition that -1 denotes the item that is
        unknown. It would be incorrect to use any other value, since 0 is a
        valid item ID.

        :return: The highest item ID.
        :rtype: int
        """
        max_val = self._df[self._df != -1][InteractionMatrix.ITEM_IX].max()
        if np.isnan(max_val):
            return -1
        return max_val

    @property
    def timestamps(self) -> pd.Series:
        """The interaction timestamps indexed by user and item ID.

        :raises
        """
        """Timestamps of interactions as a pandas Series, indexed by user ID and item ID.

        :raises TimestampAttributeMissingError: If timestamp column is missing.
        :return: Interactions with composite index on (user ID, item ID)
        :rtype: pd.Series
        """
        if not self.has_timestamps:
            raise TimestampAttributeMissingError()
        index = pd.MultiIndex.from_frame(self._df[[InteractionMatrix.USER_IX, InteractionMatrix.ITEM_IX]])
        return pd.DataFrame(self._df[[InteractionMatrix.TIMESTAMP_IX]]).set_index(index)[InteractionMatrix.TIMESTAMP_IX]

    @property
    def latest_interaction_timestamps_matrix(self) -> csr_matrix:
        """A csr matrix containing the last interaction timestamp for each user, item pair.

        We only account for the last interacted timestamp making the dataset non-deduplicated.
        """
        timestamps = self.timestamps.groupby(["uid", "iid"]).max().reset_index()
        timestamp_mat = csr_matrix(
            (timestamps.ts.values, (timestamps.uid.values, timestamps.iid.values)),
            shape=self.user_item_shape,
        )

        return timestamp_mat



# RecPack, An Experimentation Toolkit for Top-N Recommendation
# Copyright (C) 2020  Froomle N.V.
# License: GNU AGPLv3 - https://gitlab.com/recpack-maintainers/recpack/-/blob/master/LICENSE
# Author:
#   Lien Michiels
#   Robin Verachtert

import numpy as np
from numpy.typing import ArrayLike


class DecayFunction:
    def __call__(self, time_distances: ArrayLike) -> ArrayLike:
        """Apply the decay.

        :param time_distances: array of distances to be decayed.
        :type time_distances: ArrayLike
        :returns: Array of event ages to which decays have been applied.
        :rtype: ArrayLike
        """
        raise NotImplementedError()


class ExponentialDecay(DecayFunction):
    """Applies exponential decay.

    For each value x in ``time_distances`` the decayed value is computed as

    .. math::

        f(x) = e^{-\\alpha * x}

    where alpha is the decay parameter.

    :param decay: Exponential decay parameter, should be in the [0, 1] interval.
    :type decay: float
    """

    @classmethod
    def validate_decay(cls, decay: float) -> None:
        """Verify if the decay parameter is in the right range for this decay function."""
        if not (0 <= decay <= 1):
            raise ValueError(f"Decay parameter = {decay} is not in the supported range: [0, 1].")

    def __init__(self, decay: float):
        self.validate_decay(decay)
        self.decay = decay

    def __call__(self, time_distances: ArrayLike) -> ArrayLike:
        """Apply the decay function.

        :param time_distances: array of distances to be decayed.
        :type time_distances: ArrayLike
        :returns: The decayed time array.
        :rtype: ArrayLike
        """

        return np.exp(-self.decay * time_distances)


class ConvexDecay(DecayFunction):
    """Applies a convex decay function.

    For each value x in the ``time_distances`` the decayed value is computed as

    .. math::

        f(x) = \\alpha^{x}

    where :math:`alpha` is the decay parameter.

    :param decay: The decay parameter, should be in the ]0, 1] interval.
    :type decay: float
    """

    @classmethod
    def validate_decay(cls, decay: float):
        """Verify if the decay parameter is in the right range for this decay function."""
        if not (0 < decay <= 1):
            raise ValueError(f"Decay parameter = {decay} is not in the supported range: ]0, 1].")

    def __init__(self, decay: float):
        self.validate_decay(decay)
        self.decay = decay

    def __call__(self, time_distances: ArrayLike):
        """Apply the decay function.

        :param time_distances: array of distances to be decayed.
        :type time_distances: ArrayLike
        :returns: The decayed time array.
        :rtype: ArrayLike
        """

        return np.power(self.decay, time_distances)


class ConcaveDecay(DecayFunction):
    """Applies a concave decay function.

    For each value x in the ``time_distances`` the decayed value is computed as

    .. math::

        f(x) = 1 - \\alpha^{1-\\frac{x}{N}}

    where :math:`alpha` is the decay parameter and :math:`N` is the ``max_distance`` parameter.

    :param decay: The decay parameter, should be in the [0, 1[ interval.
    :type decay: float
    :param max_distance: Normalizing parameter, to put distances in the [0, 1].
    :type max_distance: float
    """

    @classmethod
    def validate_decay(cls, decay: float):
        """Verify if the decay parameter is in the right range for this decay function."""
        if not (0 < decay <= 1):
            raise ValueError(f"Decay parameter = {decay} is not in the supported range: ]0, 1].")

    def __init__(self, decay: float, max_distance: float):
        self.validate_decay(decay)
        self.decay = decay
        self.max_distance = max_distance

    def __call__(self, time_distances: ArrayLike):
        """Apply the decay function.

        :param time_distances: array of distances to be decayed.
        :type time_distances: ArrayLike
        :returns: The decayed array.
        :rtype: ArrayLike
        """
        if (time_distances > self.max_distance).any():
            raise ValueError(
                "At least one of the distances is bigger than the specified max_distance."
            )
        return 1 - np.power(self.decay, 1 - (time_distances / self.max_distance))


class LogDecay(DecayFunction):
    """Applies a logarithmic decay function.

    For each value x in the ``time_distances`` the decayed value is computed as

    .. math::

        f(x) = log_\\alpha ((\\alpha-1)(1-\\frac{x}{N}) + 1)

    where :math:`alpha` is the decay parameter and :math:`N` is the ``max_distance`` parameter.

    :param decay: The decay parameter, should be in the range ]1, inf[
    :type decay: float
    :param max_distance: Normalizing parameter, to put distances in the [0, 1].
    :type max_distance: float
    """

    @classmethod
    def validate_decay(cls, decay: float):
        """Verify if the decay parameter is in the right range for this decay function."""
        if not (1 < decay):
            raise ValueError(f"Decay parameter = {decay} is not in the supported range: ]1, inf[.")

    def __init__(self, decay: float, max_distance: float):
        self.validate_decay(decay)
        self.decay = decay
        self.max_distance = max_distance

    def __call__(self, time_distances: ArrayLike):
        """Apply the decay function.

        :param time_distances: array of distances to be decayed.
        :type time_distances: ArrayLike
        :returns: The decayed time array.
        :rtype: ArrayLike
        """
        if (time_distances > self.max_distance).any():
            raise ValueError(
                "At least one of the distances is bigger than the specified max_distance."
            )
        return np.log(((self.decay - 1) * (1 - time_distances / self.max_distance)) + 1) / np.log(
            self.decay
        )


class LinearDecay(DecayFunction):
    """Applies a linear decay function.

    For each value x in the ``time_distances`` the decayed value is computed as

    .. math::

        f(x) = \\max(1 - (\\frac{x}{N}) \\alpha, 0)

    where :math:`alpha` is the decay parameter and :math:`N` is the ``max_distance`` parameter.

    :param decay: The decay parameter, should be in the [0, inf[ interval.
    :type decay: float
    :param max_distance: Normalizing parameter, to put distances in the [0, 1].
    :type max_distance: float
    """

    @classmethod
    def validate_decay(cls, decay: float):
        if not (0 <= decay):
            raise ValueError(f"Decay parameter = {decay} is not in the supported range: [0, +inf[.")

    def __init__(self, decay: float, max_distance: float):
        self.validate_decay(decay)
        self.decay = decay
        self.max_distance = max_distance

    def __call__(self, time_distances: ArrayLike):
        """Apply the decay function.

        :param time_distances: array of distances to be decayed.
        :type time_distances: ArrayLike
        :returns: The decayed time array.
        :rtype: ArrayLike
        """
        if (time_distances > self.max_distance).any():
            raise ValueError(
                "At least one of the distances is bigger than the specified max_distance."
            )
        results = 1 - (time_distances / self.max_distance) * self.decay
        results[results < 0] = 0
        return results


class InverseDecay(DecayFunction):
    """Invert the scores.

    Decay parameter only added for interface unity.
    For each value x in the ``time_distances`` the decayed value is computed as

    .. math::

        f(x) = \\frac{1}{x}
    """

    def __call__(self, time_distances: ArrayLike):
        """Apply the decay function.

        :param time_distances: array of distances to be decayed.
        :type time_distances: ArrayLike
        :returns: The decayed time array.
        :rtype: ArrayLike
        """

        results = time_distances.astype(float).copy()
        results[results > 0] = 1 / results[results > 0]
        results[results == 0] = 1
        return results


class NoDecay(ExponentialDecay):
    """Turns the array into a binary array."""

    def __init__(self):
        super().__init__(0)


"""Data filtering module.

This module provides abstract base class and filter implementations for
removing interactions from a DataFrame based on various criteria.
"""

from abc import ABC, abstractmethod

import pandas as pd


class Filter(ABC):
    """Abstract base class for filter implementations.

    A filter must implement an `apply` method that takes a pandas DataFrame
    as input and returns a processed pandas DataFrame as output.
    """

    @abstractmethod
    def apply(self, df: pd.DataFrame) -> pd.DataFrame:
        """Apply filter to the DataFrame.

        Args:
            df: DataFrame to filter.

        Returns:
            Filtered DataFrame.
        """
        raise NotImplementedError

    def __str__(self) -> str:
        attrs = self.__dict__
        return f"{self.__class__.__name__}({', '.join((f'{k}={v}' for k, v in attrs.items()))})"


class MinItemsPerUser(Filter):
    """Filter requiring users to have minimum interaction count.

    Removes users who have interacted with fewer than the specified minimum
    number of items. Adapted from RecPack.

    Args:
        min_items_per_user: Minimum number of items a user must interact with.
        item_ix: Column name containing item identifiers.
        user_ix: Column name containing user identifiers.
        count_duplicates: Whether to count multiple interactions with the same
            item. Defaults to True.

    Example:
        Original interactions:
        ```
        user | item
        1 | a
        1 | b
        1 | c
        2 | a
        2 | b
        2 | d
        3 | a
        3 | b
        3 | d
        ```

        After `MinItemsPerUser(3)`:
        ```
        user | item
        1 | a
        1 | b
        2 | a
        2 | b
        3 | a
        3 | b
        ```

        Users 1 and 2 are removed (have fewer than 3 items).
    """

    def __init__(
        self,
        min_items_per_user: int,
        item_ix: str,
        user_ix: str,
        count_duplicates: bool = True,
    ) -> None:
        self.min_items_per_user = min_items_per_user
        self.count_duplicates = count_duplicates
        self.item_ix = item_ix
        self.user_ix = user_ix

    def apply(self, df: pd.DataFrame) -> pd.DataFrame:
        """Apply minimum items per user filter.

        Args:
            df: DataFrame to filter.

        Returns:
            DataFrame containing only users with sufficient interactions.
        """
        uids = (
            df[self.user_ix]
            if self.count_duplicates
            else df.drop_duplicates([self.user_ix, self.item_ix])[self.user_ix]
        )

        cnt_items_per_user = uids.value_counts()
        users_of_interest = list(cnt_items_per_user[cnt_items_per_user >= self.min_items_per_user].index)
        return df[df[self.user_ix].isin(users_of_interest)].copy()


class MinUsersPerItem(Filter):
    """Filter requiring items to have minimum user interaction count.

    Removes items that have been interacted with by fewer than the specified
    minimum number of users. Adapted from RecPack.

    Args:
        min_users_per_item: Minimum number of users that must interact with item.
        item_ix: Column name containing item identifiers.
        user_ix: Column name containing user identifiers.
        count_duplicates: Whether to count multiple interactions from the same
            user. Defaults to True.

    Example:
        Original interactions:
        ```
        user | item
        1 | a
        1 | b
        1 | c
        2 | a
        2 | b
        2 | d
        3 | a
        3 | b
        3 | d
        ```

        After `MinUsersPerItem(3)`:
        ```
        user | item
        1 | a
        1 | b
        2 | a
        2 | b
        3 | a
        3 | b
        ```

        Items with fewer than 3 users are removed (c and d).
    """

    def __init__(
        self,
        min_users_per_item: int,
        item_ix: str,
        user_ix: str,
        count_duplicates: bool = True,
    ) -> None:
        self.item_ix = item_ix
        self.user_ix = user_ix
        self.min_users_per_item = min_users_per_item
        self.count_duplicates = count_duplicates

    def apply(self, df: pd.DataFrame) -> pd.DataFrame:
        """Apply minimum users per item filter.

        Args:
            df: DataFrame to filter.

        Returns:
            DataFrame containing only items with sufficient user interactions.
        """
        iids = (
            df[self.item_ix]
            if self.count_duplicates
            else df.drop_duplicates([self.user_ix, self.item_ix])[self.item_ix]
        )

        cnt_users_per_item = iids.value_counts()
        items_of_interest = list(cnt_users_per_item[cnt_users_per_item >= self.min_users_per_item].index)
        return df[df[self.item_ix].isin(items_of_interest)].copy()



"""
.. currentmodule:: recnexteval.algorithms

The algorithms module in recnexteval contains a collection of baseline algorithms
and various of the item-based KNN collaborative filtering algorithm. A total of
3 variation of the item-based KNN algorithm is implemented in the module. Which
are listed below

Algorithm
---------
Base class for all algorithms. Programmer should inherit from this class when
implementing a new algorithm. It provides a common interface for all algorithms
such that the expected methods and properties are defined to avoid any runtime
errors.

.. autosummary::
    :toctree: generated/

    Algorithm

Baseline Algorithms
-------------------

The baseline algorithms are simple algorithms that can be used as a reference
point to compare the performance of the more complex algorithms. The following
baseline algorithms are implemented in the module.

.. autosummary::
    :toctree: generated/

    Random
    RecentPopularity
    DecayPopularity
    MostPopular

Item Similarity Algorithms
----------------------------

Item similarity algorithms exploit relationships between items to make recommendations.
At prediction time, the user is represented by the items they have interacted
with. 3 variations of the item-based KNN algorithm are implemented in the module.
Each variation is to showcase the difference in the learning and prediction of
the algorithm. We note that no one algorithm is better than the other, and it
greatly depends on the dataset and parameters used in the algorithm which would
yield the best performance.

.. autosummary::
    :toctree: generated/

    ItemKNN
    ItemKNNIncremental
    ItemKNNIncrementalMovieLens100K
    ItemKNNRolling
    ItemKNNStatic

Time-Aware Item Similarity Algorithms
-------------------------------------

This is an extension of the item-based KNN algorithm.
The time-aware item-based KNN algorithms applies a decay factor to the timestamps of interactions.
This gives heavier weight to interactions that are more recent.

.. autosummary::
    :toctree: generated/

    TARSItemKNN
    TARSItemKNNLiu
    TARSItemKNNVaz
    TARSItemKNNDing
"""

from .base import Algorithm
from .baseline import MostPopular, Random, RecentPopularity
from .baseline.decay_popularity import DecayPopularity
from .itemknn import ItemKNN, ItemKNNIncremental, ItemKNNIncrementalMovieLens100K, ItemKNNRolling, ItemKNNStatic
from .time_aware_item_knn import (
    TARSItemKNN,
    TARSItemKNNDing,
    TARSItemKNNLiu,
    TARSItemKNNVaz,
)


__all__ = [
    "Algorithm",
    "DecayPopularity",
    "ItemKNN",
    "ItemKNNIncremental",
    "ItemKNNIncrementalMovieLens100K",
    "ItemKNNRolling",
    "ItemKNNStatic",
    "MostPopular",
    "Random",
    "RecentPopularity",
    "TARSItemKNN",
    "TARSItemKNNDing",
    "TARSItemKNNLiu",
    "TARSItemKNNVaz",
]



import logging

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import OneHotEncoder

from ...matrix import InteractionMatrix
from ...utils import add_rows_to_csr_matrix
from .itemknn_incremental import ItemKNNIncremental


logger = logging.getLogger(__name__)


class ItemKNNIncrementalMovieLens100K(ItemKNNIncremental):
    """Incremental version of ItemKNN algorithm with MovieLens100k Metadata.

    This class extends the ItemKNN algorithm to allow for incremental updates
    to the model. The incremental updates are done by updating the historical
    data with the new data by appending the new data to the historical data.
    """
    IS_BASE: bool = False

    def __init__(self, metadata: pd.DataFrame, K:int=10) -> None:
        super().__init__(K)
        if metadata is None:
            raise ValueError("Metadata is required for ItemKNNIncrementalMovieLens100K")
        self.metadata = metadata.copy()

    def _predict(self, X: csr_matrix, predict_im: InteractionMatrix) -> csr_matrix:
        """Predict the K most similar items for each item using the latest data."""
        X_pred = super()._predict(self.X_)
        # ID indexing starts at 0, so max_id + 1 is the number of unique IDs
        max_user_id = predict_im.max_user_id + 1
        max_item_id = predict_im.max_item_id + 1
        intended_shape = (
            max(max_user_id, X.shape[0]),
            max(max_item_id, X.shape[1]),
        )

        predict_frame = predict_im._df

        if X_pred.shape == intended_shape:
            return X_pred

        known_user_id, known_item_id = X_pred.shape
        X_pred = add_rows_to_csr_matrix(X_pred, intended_shape[0] - known_user_id)
        logger.debug(f"Padding user ID in range({known_user_id}, {intended_shape[0]}) with items")
        to_predict = predict_frame.value_counts("uid")

        # pad users with items from most similar user
        user_similarity_matrix = self.get_user_similarity_matrix()
        for user_id in to_predict.index:
            if user_id >= known_user_id:
                most_similar_user_idx = np.argmax(user_similarity_matrix[user_id][:known_user_id])
                X_pred[user_id, :] = X_pred[most_similar_user_idx, :]

        logger.debug(f"Padding by {self.name} completed")
        return X_pred

    def get_user_similarity_matrix(self):
        user_metadata = self.metadata.copy()

        # set userId as index
        user_metadata.set_index("userId", inplace=True)
        user_metadata.index.name = None

        # reorder the indices
        user_metadata.reset_index(drop=True)
        user_metadata.sort_index(inplace=True)

        # zipcode is a column that does not provide any useful information so we drop it
        user_metadata = user_metadata.drop(columns=["zipcode"])

        # obtain categorical columns
        categorical_columns = user_metadata.select_dtypes(include=["object"]).columns.tolist()

        # Use one-hot encoding to encode the categorical columns
        encoder = OneHotEncoder(sparse_output=False)
        one_hot_encoded = encoder.fit_transform(user_metadata[categorical_columns])

        # obtain the column names for the encoded data
        one_hot_df = pd.DataFrame(one_hot_encoded, columns=encoder.get_feature_names_out(categorical_columns))

        # Concatenate the one-hot encoded dataframe with the original dataframe and drop the original categorical columns
        df_encoded = pd.concat([user_metadata, one_hot_df], axis=1)
        df_encoded = df_encoded.drop(categorical_columns, axis=1)

        # compute cosine similarity but exclude self-similarity
        user_similarity_matrix = cosine_similarity(df_encoded)
        np.fill_diagonal(user_similarity_matrix, 0)

        return user_similarity_matrix



import logging
from typing import Self

from recnexteval.matrix import InteractionMatrix
from .itemknn import ItemKNN


logger = logging.getLogger(__name__)


class ItemKNNStatic(ItemKNN):
    """Static version of ItemKNN algorithm.

    This class extends the ItemKNN algorithm to only fit the model once. `fit` will only
    fit the model once and will not update the model with new data. The purpose
    is to make the training data static and not update the model with new data.
    """

    IS_BASE: bool = False

    def __init__(self, K: int = 10) -> None:
        self._is_fitted = False
        super().__init__(K)

    def fit(self, X: InteractionMatrix) -> Self:
        if self._is_fitted:
            return self

        super().fit(X)
        self._is_fitted = True
        return self



from scipy.sparse import csr_matrix

from recnexteval.matrix.interaction_matrix import InteractionMatrix
from recnexteval.utils.util import to_binary


def to_csr_matrix(X: InteractionMatrix | csr_matrix, binary: bool = False) -> csr_matrix:
    """Convert a matrix-like object to a scipy csr_matrix.

    :param X: Matrix-like object or tuple of objects to convert.
    :type X: csr_matrix
    :param binary: If true, ensure matrix is binary by setting non-zero values to 1.
    :type binary: bool, optional
    :return: Matrices as csr_matrix.
    :rtype: Union[csr_matrix, Tuple[csr_matrix, ...]]
    """

    if isinstance(X, csr_matrix):
        res = X
    elif isinstance(X, InteractionMatrix):
        res = X.values
    else:
        raise AttributeError("Not supported Matrix conversion")
    return to_binary(res) if binary else res



import logging
from enum import Enum
from uuid import UUID

from scipy.sparse import csr_matrix

from recnexteval.algorithms import Algorithm
from recnexteval.matrix import InteractionMatrix, PredictionMatrix
from recnexteval.registries import (
    METRIC_REGISTRY,
    MetricEntry,
)
from recnexteval.settings import EOWSettingError, Setting
from .accumulator import MetricAccumulator
from .base import EvaluatorBase
from .state_management import AlgorithmStateEnum, AlgorithmStateManager
from .strategy import EvaluationStrategy, SlidingWindowStrategy


class EvaluatorState(Enum):
    """Evaluator lifecycle states"""

    INITIALIZED = "initialized"
    STARTED = "started"
    IN_PROGRESS = "in_progress"
    COMPLETED = "completed"
    FAILED = "failed"


logger = logging.getLogger(__name__)


class EvaluatorStreamer(EvaluatorBase):
    """Evaluation via streaming through API.

    The diagram below shows the diagram of the streamer evaluator for the
    sliding window setting. Instead of the pipeline, we allow the user to
    stream the data release to the algorithm. The data communication is shown
    between the evaluator and the algorithm. Note that while only 2 splits are
    shown here, the evaluator will continue to stream the data until the end
    of the setting where there are no more splits.

    ![stream scheme](../../../assets/_static/stream_scheme.png)

    This class exposes a few of the core API that allows the user to stream
    the evaluation process. The following API are exposed:

    1. :meth:`register_algorithm`
    2. :meth:`start_stream`
    3. :meth:`get_unlabeled_data`
    4. :meth:`submit_prediction`

    The programmer can take a look at the specific method for more details
    on the implementation of the API. The methods are designed with the
    methodological approach that the algorithm is decoupled from the
    the evaluating platform. And thus, the evaluator will only provide
    the necessary data to the algorithm and evaluate the prediction.

    Args:
        metric_entries: list of metric entries.
        setting: Setting object.
        metric_k: Number of top interactions to consider.
        ignore_unknown_user: To ignore unknown users.
        ignore_unknown_item: To ignore unknown items.
        seed: Random seed for the evaluator.
    """

    def __init__(
        self,
        metric_entries: list[MetricEntry],
        setting: Setting,
        metric_k: int,
        ignore_unknown_user: bool = False,
        ignore_unknown_item: bool = False,
        seed: int = 42,
        strategy: None | EvaluationStrategy = None,
    ) -> None:
        super().__init__(
            metric_entries,
            setting,
            metric_k,
            ignore_unknown_user,
            ignore_unknown_item,
            seed,
        )
        self._algo_state_mgr = AlgorithmStateManager()
        self._unlabeled_data_cache: PredictionMatrix
        self._ground_truth_data_cache: PredictionMatrix
        self._training_data_cache: PredictionMatrix

        # Evaluator state management
        self._state = EvaluatorState.INITIALIZED

        # Evaluation strategy
        self._strategy = strategy or SlidingWindowStrategy()

    @property
    def state(self) -> EvaluatorState:
        return self._state

    def _assert_state(self, expected: EvaluatorState, error_msg: str) -> None:
        """Assert evaluator is in expected state"""
        if self._state != expected:
            raise RuntimeError(f"{error_msg} (Current state: {self._state.value})")

    def _transition_state(self, new_state: EvaluatorState, allow_from: list[EvaluatorState]) -> None:
        """Guard state transitions explicitly"""
        if self._state not in allow_from:
            raise ValueError(f"Cannot transition from {self._state} to {new_state}. Allowed from: {allow_from}")
        self._state = new_state
        logger.info(f"Evaluator transitioned to {new_state}")

    def _cache_evaluation_data(self) -> None:
        """Cache the evaluation data for the current step.

        Summary
        -------
        This method will cache the evaluation data for the current step. The method
        will update the unknown user/item base, get the next unlabeled and ground
        truth data, and update the current timestamp.

        Specifics
        ---------
        The method will update the unknown user/item base with the ground truth data.
        Next, mask the unlabeled and ground truth data with the known user/item
        base. The method will cache the unlabeled and ground truth data in the internal
        attributes :attr:`_unlabeled_data_cache` and :attr:`_ground_truth_data_cache`.
        The timestamp is cached in the internal attribute :attr:`_current_timestamp`.

        We use an internal attribute :attr:`_run_step` to keep track of the current
        step such that we can check if we have reached the last step.

        We assume that any method calling this method has already checked if the
        there is still data to be processed.
        """

        logger.debug(f"Caching evaluation data for step {self._run_step}")
        try:
            self._unlabeled_data_cache, self._ground_truth_data_cache, _ = self._get_evaluation_data()
        except EOWSettingError as e:
            raise e
        logger.debug(f"Data cached for step {self._run_step} complete")

    def start_stream(self) -> None:
        """Start the streaming process.

        This method is called to start the streaming process. `start_stream` will
        prepare the evaluator for the streaming process. `start_stream` will reset
        data streamers, prepare the micro and macro accumulators, update
        the known user/item base, and cache data.

        The method will set the internal state to be be started. The
        method can be called anytime after the evaluator is instantiated.

        Warning:
            Once `start_stream` is called, the evaluator cannot register any new algorithms.

        Raises:
            ValueError: If the stream has already started.
        """
        self.setting.restore()

        logger.debug("Preparing evaluator for streaming")
        self._acc = MetricAccumulator()
        training_data = self.setting.background_data
        # Convert to PredictionMatrix since it's a subclass of InteractionMatrix
        training_data = PredictionMatrix.from_interaction_matrix(training_data)

        self.user_item_base.update_known_user_item_base(training_data)
        training_data.mask_user_item_shape(self.user_item_base.known_shape)
        self._training_data_cache = training_data
        self._cache_evaluation_data()
        self._algo_state_mgr.set_all_ready(data_segment=self._current_timestamp)
        logger.debug("Evaluator is ready for streaming")
        # TODO: allow programmer to register anytime
        self._transition_state(EvaluatorState.STARTED, allow_from=[EvaluatorState.INITIALIZED])

    def register_algorithm(
        self,
        algorithm: None | Algorithm = None,
        algorithm_name: None | str = None,
    ) -> UUID:
        """Register the algorithm with the evaluator.

        This method is called to register the algorithm with the evaluator.
        The method will assign a unique identifier to the algorithm and store
        the algorithm in the registry. The method will raise a ValueError if
        the stream has already started.
        """
        self._assert_state(EvaluatorState.INITIALIZED, "Cannot register algorithms after stream started")
        algo_id = self._algo_state_mgr.register(name=algorithm_name, algo_ptr=algorithm)
        logger.debug(f"Algorithm {algo_id} registered")
        return algo_id

    def get_algorithm_state(self, algo_id: UUID) -> AlgorithmStateEnum:
        """Get the state of the algorithm.

        This method is called to get the state of the algorithm given the
        unique identifier of the algorithm. The method will return the state
        of the algorithm.

        Args:
            algo_id: Unique identifier of the algorithm.

        Returns:
            The state of the algorithm.
        """
        return self._algo_state_mgr[algo_id].state

    def get_all_algorithm_status(self) -> dict[str, AlgorithmStateEnum]:
        """Get the status of all algorithms.

        This method is called to get the status of all algorithms registered
        with the evaluator. The method will return a dictionary where the key
        is the name of the algorithm and the value is the state of the algorithm.

        Returns:
            The status of all algorithms.
        """
        return self._algo_state_mgr.all_algo_states()

    def load_next_window(self) -> None:
        self.user_item_base.reset_unknown_user_item_base()
        incremental_data = self.setting.get_split_at(self._run_step).incremental
        if incremental_data is None:
            raise EOWSettingError("No more data to stream")
        # Convert to PredictionMatrix since it's a subclass of InteractionMatrix
        incremental_data = PredictionMatrix.from_interaction_matrix(incremental_data)

        self.user_item_base.update_known_user_item_base(incremental_data)
        incremental_data.mask_user_item_shape(self.user_item_base.known_shape)
        self._training_data_cache = incremental_data
        self._cache_evaluation_data()
        self._algo_state_mgr.set_all_ready(data_segment=self._current_timestamp)

    def get_training_data(self, algo_id: UUID) -> InteractionMatrix:
        """Get training data for the algorithm.

        Summary
        -------

        This method is called to get the training data for the algorithm. The
        training data is defined as either the background data or the incremental
        data. The training data is always released irrespective of the state of
        the algorithm.

        Specifics
        ---------

        1. If the state is COMPLETED, raise warning that the algorithm has completed
        2. If the state is NEW, release training data to the algorithm
        3. If the state is READY and the data segment is the same, raise warning
           that the algorithm has already obtained data
        4. If the state is PREDICTED and the data segment is the same, inform
           the algorithm that it has already predicted and should wait for other
           algorithms to predict
        5. This will occur when :attr:`_current_timestamp` has changed, which causes
           scenario 2 to not be caught. In this case, the algorithm is requesting
           the next window of data. Thus, this is a valid data call and the status
           will be updated to READY.

        Args:
            algo_id: Unique identifier of the algorithm.

        Raises:
            ValueError: If the stream has not started.

        Returns:
            The training data for the algorithm.
        """
        self._assert_state(EvaluatorState.STARTED, "Call start_stream() first")

        logger.debug(f"Getting data for algorithm {algo_id}")

        if self._strategy.should_advance_window(
            algo_state_mgr=self._algo_state_mgr,
            current_step=self._run_step,
            total_steps=self.setting.num_split,
        ):
            try:
                self.load_next_window()
            except EOWSettingError:
                self._transition_state(
                    EvaluatorState.COMPLETED, allow_from=[EvaluatorState.STARTED, EvaluatorState.IN_PROGRESS]
                )
                raise RuntimeError("End of evaluation window reached")

        can_request, reason = self._algo_state_mgr.can_request_training_data(algo_id)
        if not can_request:
            raise PermissionError(f"Cannot request data: {reason}")
        # TODO handle case when algo is ready after submitting prediction, but current timestamp has not changed, meaning algo is requesting same data again
        self._algo_state_mgr.transition(
            algo_id,
            AlgorithmStateEnum.RUNNING,
            data_segment=self._current_timestamp,
        )

        self._evaluator_state = EvaluatorState.IN_PROGRESS
        # release data to the algorithm
        return self._training_data_cache

    def get_unlabeled_data(self, algo_id: UUID) -> PredictionMatrix:
        """Get unlabeled data for the algorithm.

        This method is called to get the unlabeled data for the algorithm. The
        unlabeled data is the data that the algorithm will predict. It will
        contain `(user_id, -1)` pairs, where the value -1 indicates that the
        item is to be predicted.
        """
        logger.debug(f"Getting unlabeled data for algorithm {algo_id}")
        can_submit, reason = self._algo_state_mgr.can_request_unlabeled_data(algo_id)
        if not can_submit:
            raise PermissionError(f"Cannot get unlabeled data: {reason}")
        return self._unlabeled_data_cache

    def submit_prediction(self, algo_id: UUID, X_pred: csr_matrix) -> None:
        """Submit the prediction of the algorithm.

        This method is called to submit the prediction of the algorithm.
        There are a few checks that are done before the prediction is
        evaluated by calling :meth:`_evaluate_algo_pred`.

        Once the prediction is evaluated, the method will update the state
        of the algorithm to PREDICTED.
        """
        logger.debug(f"Submitting prediction for algorithm {algo_id}")
        can_submit, reason = self._algo_state_mgr.can_submit_prediction(algo_id)
        if not can_submit:
            raise PermissionError(f"Cannot submit prediction: {reason}")

        self._evaluate_algo_pred(algo_id=algo_id, y_pred=X_pred)
        self._algo_state_mgr.transition(
            algo_id,
            AlgorithmStateEnum.PREDICTED,
        )

    def _evaluate_algo_pred(self, algo_id: UUID, y_pred: csr_matrix) -> None:
        """Evaluate the prediction for algorithm.

        Given the prediction and the algorithm ID, the method will evaluate the
        prediction using the metrics specified in the evaluator. The prediction
        of the algorithm is compared to the ground truth data currently cached.

        The evaluation results will be stored in the micro and macro accumulators
        which will later be used to calculate the final evaluation results.

        Args:
            algo_id: The unique identifier of the algorithm.
            y_pred: The prediction of the algorithm.
        """
        # get top k ground truth interactions
        y_true = self._ground_truth_data_cache
        # y_true = self._ground_truth_data_cache.get_users_n_first_interaction(self.metric_k)
        y_true = y_true.item_interaction_sequence_matrix

        y_pred = self._prediction_shape_handler(y_true, y_pred)
        algorithm_name = self._algo_state_mgr.get_algorithm_identifier(algo_id)

        # evaluate the prediction
        for metric_entry in self.metric_entries:
            metric_cls = METRIC_REGISTRY.get(metric_entry.name)
            params = {
                'timestamp_limit': self._current_timestamp,
                'user_id_sequence_array': self._ground_truth_data_cache.user_id_sequence_array,
                'user_item_shape': self._ground_truth_data_cache.user_item_shape,
            }
            if metric_entry.K is not None:
                params['K'] = metric_entry.K

            metric = metric_cls(**params)
            metric.calculate(y_true, y_pred)
            self._acc.add(metric=metric, algorithm_name=algorithm_name)

        logger.debug(f"Prediction evaluated for algorithm {algo_id} complete")



"""
RecNextEval
-----------

RecNextEval is a Python package toolkit developed for evaluation of recommendation
systems in different settings. Mainly the toolkit is developed to evaluate
in a sliding window setting.
"""

import logging

from recnexteval.utils import prepare_logger


LOGGING_CONFIG_FILENAME = "logging_config.yaml"

prepare_logger(LOGGING_CONFIG_FILENAME)

logger = logging.getLogger(__name__)
logger.info("recnexteval package loaded.")



import logging
from warnings import warn

import numpy as np

from recnexteval.matrix import InteractionMatrix, TimestampAttributeMissingError
from .base import Setting
from .splitters import (
    NLastInteractionTimestampSplitter,
    TimestampSplitter,
)


logger = logging.getLogger(__name__)


class SingleTimePointSetting(Setting):
    """Single time point setting for data split.

    Splits an interaction dataset at a single timestamp into background
    (training) data and evaluation data. The evaluation data can be
    further processed to produce unlabeled inputs and ground-truth
    targets for model evaluation.

    Args:
        background_t: Time point to split the data. The background
            split covers interactions with timestamps in `[0, background_t)`.
        n_seq_data: Number of last sequential interactions
            to provide as input for prediction. Defaults to `1`.
        top_K: Number of interactions per user to select for
            evaluation purposes. Defaults to `1`.
        t_upper: Upper bound on the timestamp of
            interactions included in evaluation. Defaults to the maximum
            32-bit integer value (acts like infinity).
        include_all_past_data: If True, include all past
            interactions when constructing input sequences. Defaults to False.
        seed: Random seed for reproducible behavior.
            If None, a seed will be generated.
    """
    IS_BASE: bool = False

    def __init__(
        self,
        background_t: int,
        n_seq_data: int = 1,
        top_K: int = 1,
        t_upper: int = np.iinfo(np.int32).max,
        include_all_past_data: bool = False,
        seed: int = 42,
    ):
        super().__init__(seed=seed)
        self.t = background_t
        """Seconds before `t` timestamp value to be used in `background_set`."""
        self.t_upper = t_upper
        """Seconds after `t` timestamp value to be used in `ground_truth_data`."""
        self.n_seq_data = n_seq_data
        self.top_K = top_K

        logger.info("Splitting data at time %s with t_upper interval %s", background_t, t_upper)

        self._background_splitter = TimestampSplitter(
            t=background_t,
            t_lower=None,
            t_upper=t_upper,
        )
        self._splitter = NLastInteractionTimestampSplitter(
            t=background_t,
            t_upper=t_upper,
            n_seq_data=n_seq_data,
            include_all_past_data=include_all_past_data,
        )
        self._t_window = background_t

    def _split(self, data: InteractionMatrix) -> None:
        """Split the dataset by timestamp into background and evaluation sets.

        The method raises :class:`TimestampAttributeMissingError` when the
        provided :class:`InteractionMatrix` does not contain timestamp
        information. It will warn if the chosen split time is before the
        earliest timestamp in the data.

        Args:
            data: Interaction matrix to split. Must have timestamps.

        Raises:
            TimestampAttributeMissingError: If `data` has no timestamp attribute.
        """
        if not data.has_timestamps:
            raise TimestampAttributeMissingError()
        if data.min_timestamp > self.t:
            warn(
                f"Splitting at time {self.t} is before the first timestamp"
                " in the data. No data will be in the training set."
            )

        self._background_data, _ = self._background_splitter.split(data)
        past_interaction, future_interaction = self._splitter.split(data)
        self._unlabeled_data, self._ground_truth_data = self.prediction_data_processor.process(
            past_interaction=past_interaction,
            future_interaction=future_interaction,
            top_K=self.top_K,
        )

        if len(self._background_data) == 0:
            logger.info("Background data is empty after splitting at time %s", self.t)
        if len(self._unlabeled_data) == 0:
            logger.info("Unlabeled data is empty after splitting at time %s", self.t)
        if len(self._ground_truth_data) == 0:
            logger.info("Ground truth data is empty after splitting at time %s", self.t)

        logger.info("Finished splitting data at time %s", self.t)



import logging
from abc import ABC, abstractmethod
from warnings import warn

from recnexteval.registries import (
    METRIC_REGISTRY,
    MetricEntry,
)
from recnexteval.settings import Setting
from recnexteval.utils import arg_to_str
from ..base import EvaluatorBase


logger = logging.getLogger(__name__)


class Builder(ABC):
    """Base class for Builder objects.

    Provides methods to set specific values for the builder and enforce checks
    such that the builder can be constructed correctly and to avoid possible
    errors when the builder is executed.
    """

    def __init__(
        self,
        ignore_unknown_user: bool = True,
        ignore_unknown_item: bool = True,
        seed: int = 42,
    ) -> None:
        """Initialize the Builder.

        Args:
            ignore_unknown_user: Ignore unknown user in the evaluation.
            ignore_unknown_item: Ignore unknown item in the evaluation.
            seed: Random seed for reproducibility.
        """
        self.metric_entries: dict[str, MetricEntry] = dict()
        """dict of metrics to evaluate algorithm on.
        Using dict instead of list for fast lookup"""
        self.setting: Setting
        """Setting to evaluate the algorithms on"""
        self.ignore_unknown_user = ignore_unknown_user
        """Ignore unknown user in the evaluation"""
        self.ignore_unknown_item = ignore_unknown_item
        """Ignore unknown item in the evaluation"""
        self.metric_k: int
        self.seed: int = seed

    def _check_setting_exist(self) -> bool:
        """Check if setting is already set.

        Returns:
            True if setting is set, False otherwise.
        """
        return not (not hasattr(self, "setting") or self.setting is None)

    def set_metric_K(self, K: int) -> None:
        """Set K value for all metrics.

        Args:
            K: K value to set for all metrics.
        """
        self.metric_k = K

    def add_metric(self, metric: str | type) -> None:
        """Add metric to evaluate algorithm on.

        Metric will be added to the metric_entries dict where it will later be
        converted to a list when the evaluator is constructed.

        Note:
            If K is not yet specified, the setting's top_K value will be used. This
            requires the setting to be set before adding the metric.

        Args:
            metric: Metric to evaluate algorithm on.

        Raises:
            ValueError: If metric is not found in METRIC_REGISTRY.
            RuntimeError: If setting is not set.
        """
        if not self._check_setting_exist():
            raise RuntimeError(
                "Setting has not been set. To ensure conformity, of the addition of"
                " other components please set the setting first. Call add_setting() method."
            )

        metric = arg_to_str(metric)

        if metric not in METRIC_REGISTRY:
            raise ValueError(f"Metric {metric} could not be resolved.")

        if not hasattr(self, "metric_k"):
            self.metric_k = self.setting.top_K
            warn(
                "K value not yet specified before setting metric, using setting's top_K value."
                " We recommend specifying K value for metric. If you want to change the K value,"
                " you can clear all metric entry and set the K value before adding metrics."
            )

        metric_name = f"{metric}_{self.metric_k}"
        if metric_name in self.metric_entries:
            logger.warning(f"Metric {metric_name} already exists. Skipping adding metric.")
            return

        self.metric_entries[metric_name] = MetricEntry(metric, self.metric_k)

    def add_setting(self, setting: Setting) -> None:
        """Add setting to the evaluator builder.

        Note:
            The setting should be set before adding metrics or algorithms
            to the evaluator.

        Args:
            setting: Setting to evaluate the algorithms on.

        Raises:
            ValueError: If setting is not of instance Setting.
        """
        if not isinstance(setting, Setting):
            raise ValueError(f"setting should be of type Setting, got {type(setting)}")
        if hasattr(self, "setting") and self.setting is not None:
            warn("Setting is already set. Continuing will overwrite the setting.")

        self.setting = setting

    def clear_metrics(self) -> None:
        """Clear all metrics from the builder."""
        self.metric_entries.clear()

    def _check_ready(self) -> None:
        """Check if the builder is ready to construct Evaluator.

        Raises:
            RuntimeError: If there are invalid configurations.
        """
        if not hasattr(self, "metric_k"):
            self.metric_k = self.setting.top_K
            warn(
                "K value not yet specified before setting metric, using setting's top_K value."
                " We recommend specifying K value for metric. If you want to change the K value,"
                " you can clear all metric entry and set the K value before adding metrics."
            )

        if len(self.metric_entries) == 0:
            raise RuntimeError("No metrics specified, can't construct Evaluator")

        # Check for settings #
        if self.setting is None:
            raise RuntimeError("No settings specified, can't construct Evaluator")
        if not self.setting.is_ready:
            raise RuntimeError(
                "Setting is not ready, can't construct Evaluator. "
                "Call split() on the setting first."
            )

    @abstractmethod
    def build(self) -> EvaluatorBase:
        """Build object.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError



import logging

from recnexteval.evaluators.evaluator_stream import EvaluatorStreamer
from .base import Builder


logger = logging.getLogger(__name__)


class EvaluatorStreamerBuilder(Builder):
    """Builder to facilitate construction of evaluator.

    Provides methods to set specific values for the evaluator and enforce checks
    such that the evaluator can be constructed correctly and to avoid possible
    errors when the evaluator is executed.
    """

    def __init__(
        self,
        ignore_unknown_user: bool = False,
        ignore_unknown_item: bool = False,
        seed: int = 42,
    ) -> None:
        """Initialize the EvaluatorStreamerBuilder.

        Args:
            ignore_unknown_user: Ignore unknown user in the evaluation.
            ignore_unknown_item: Ignore unknown item in the evaluation.
            seed: Random seed for reproducibility.
        """
        super().__init__(
            ignore_unknown_user=ignore_unknown_user,
            ignore_unknown_item=ignore_unknown_item,
            seed=seed,
        )

    def build(self) -> EvaluatorStreamer:
        """Build Evaluator object.

        Raises:
            RuntimeError: If no metrics, algorithms or settings are specified.

        Returns:
            EvaluatorStreamer: The built evaluator object.
        """
        self._check_ready()
        return EvaluatorStreamer(
            metric_entries=list(self.metric_entries.values()),
            setting=self.setting,
            metric_k=self.metric_k,
            ignore_unknown_user=self.ignore_unknown_user,
            ignore_unknown_item=self.ignore_unknown_item,
            seed=self.seed,
        )



"""Metadata module for dataset information.

This module allows users to include metadata information corresponding to datasets.
Metadata classes are built on top of the `Metadata` base class, allowing for easy
extension and customization.

## Important Notes

User and item IDs in the metadata module are mapped according to RecNextEval's
internal mapping, not the original IDs. Developers should not load metadata from
source separately. Instead, implement the metadata class and load metadata while
loading the dataset.

## Available Metadata

- `Metadata`: Abstract base class for metadata implementations
- `MovieLens100kUserMetadata`: User metadata from MovieLens 100K dataset
- `MovieLens100kItemMetadata`: Item metadata from MovieLens 100K dataset
- `AmazonBookItemMetadata`: Item metadata from Amazon Books dataset
- `AmazonMovieItemMetadata`: Item metadata from Amazon Movies dataset
- `AmazonMusicItemMetadata`: Item metadata from Amazon Music dataset
- `LastFMUserMetadata`: User metadata from Last.FM dataset
- `LastFMItemMetadata`: Item metadata from Last.FM dataset
- `LastFMTagMetadata`: Tag metadata from Last.FM dataset

## Example

Load metadata from the MovieLens 100K dataset:

```python
from recnexteval.datasets.movielens import MovieLens100K

dataset = MovieLens100K(fetch_dataset=True)
data = dataset.load()
```
"""

from .amazon import (
    AmazonBookItemMetadata,
    AmazonMovieItemMetadata,
    AmazonMusicItemMetadata,
)
from .base import Metadata
from .lastfm import LastFMItemMetadata, LastFMTagMetadata, LastFMUserMetadata
from .movielens import MovieLens100kItemMetadata, MovieLens100kUserMetadata


__all__ = [
    "Metadata",
    "MovieLens100kUserMetadata",
    "MovieLens100kItemMetadata",
    "AmazonBookItemMetadata",
    "AmazonMovieItemMetadata",
    "AmazonMusicItemMetadata",
    "LastFMUserMetadata",
    "LastFMItemMetadata",
    "LastFMTagMetadata",
]


import logging
import os
from typing import ClassVar

import pandas as pd

from recnexteval.datasets.config import (
    AmazonBooksItemMetadataConfig,
    AmazonDigitalMusicItemMetadataConfig,
    AmazonItemMetadataConfig,
    AmazonMoviesAndTVItemMetadataConfig,
    AmazonSubscriptionBoxesItemMetadataConfig,
)
from .base import Metadata


logger = logging.getLogger(__name__)


class AmazonItemMetadata(Metadata):
    config: ClassVar[AmazonItemMetadataConfig] = AmazonItemMetadataConfig()

    def __init__(self, item_id_mapping: pd.DataFrame) -> None:
        super().__init__()
        self.item_id_mapping = item_id_mapping

    def _load_dataframe(self) -> pd.DataFrame:
        self.fetch_dataset()
        df = pd.read_json(
            self.file_path,  # Ensure file_path contains the JSONL file path
            dtype=self.config.dtype_dict,
            lines=True,  # Required for JSONL format
        )

        item_id_to_iid = dict(zip(self.item_id_mapping[self.config.item_ix], self.item_id_mapping["iid"]))

        # Map config.item_ix in metadata_df using the optimized function
        df[self.config.item_ix] = df[self.config.item_ix].map(lambda x: item_id_to_iid.get(x, x))

        return df

    def _download_dataset(self) -> None:
        """Downloads the metadata for the dataset.

        Downloads the zipfile, and extracts the ratings file to `self.file_path`
        """
        if not self.config.dataset_url:
            raise ValueError(f"{self.name} does not have URL specified.")

        self._fetch_remote(
            self.config.dataset_url, os.path.join(self.base_path, f"{self.config.remote_filename}")
        )


class AmazonMusicItemMetadata(AmazonItemMetadata):
    config: ClassVar[AmazonDigitalMusicItemMetadataConfig] = AmazonDigitalMusicItemMetadataConfig()


class AmazonMovieItemMetadata(AmazonItemMetadata):
    config: ClassVar[AmazonMoviesAndTVItemMetadataConfig] = AmazonMoviesAndTVItemMetadataConfig()


class AmazonSubscriptionBoxesItemMetadata(AmazonItemMetadata):
    config: ClassVar[AmazonSubscriptionBoxesItemMetadataConfig] = AmazonSubscriptionBoxesItemMetadataConfig()


class AmazonBookItemMetadata(AmazonItemMetadata):
    config: ClassVar[AmazonBooksItemMetadataConfig] = AmazonBooksItemMetadataConfig()



from dataclasses import dataclass

from recnexteval.utils import get_data_dir


@dataclass
class DatasetConfig:
    """Base configuration for datasets."""

    user_ix: str = "user_id"
    """Name of the column in the DataFrame with user identifiers"""
    item_ix: str = "item_id"
    """Name of the column in the DataFrame with item identifiers"""
    timestamp_ix: str = "timestamp"
    """Name of the column in the DataFrame that contains time of interaction in seconds since epoch."""
    dataset_url: str = "http://example.com"
    """URL to fetch the dataset from."""
    default_base_path: str = str(get_data_dir())
    """Default base path where the dataset will be stored."""
    remote_zipname: str = ""
    remote_filename: str = ""
    """Name of the file containing user interaction."""

    @property
    def default_filename(self) -> str:
        """Derived filename from remote components."""
        if not self.remote_zipname or not self.remote_filename:
            return "dataset.csv"
        return f"{self.remote_zipname}_{self.remote_filename}"


@dataclass
class MetadataConfig(DatasetConfig):
    sep: str = "|"
    """Column separator in the data file."""

    def __post_init__(self) -> None:
        self.default_base_path = super().default_base_path + "/metadata"

    @property
    def column_names(self) -> list[str]:
        """
        Ordered list of column names for pd.read_table.

        Returns:
            list[str]: Column names in file order [user_id, age, gender, ...]

        Example:
            ["userId", "age", "gender", "occupation", "zipcode"]
        """
        return []

    @property
    def dtype_dict(self) -> dict:
        """
        Data type mapping for all columns.

        Used in pd.read_table() dtype parameter to ensure correct
        column types are loaded from file.

        Returns:
            dict: Mapping of column names to numpy dtypes

        Example:
            {
                "age": "int64",
                "gender": "<U1",  # string
                "occupation": "object",
                "zipcode": "object"
            }
        """
        return {}



import logging
from typing import ClassVar

import numpy as np
import pandas as pd

from ..config import (
    AmazonBookDatasetConfig,
    AmazonDatasetConfig,
    AmazonMovieDatasetConfig,
    AmazonMusicDatasetConfig,
    AmazonSubscriptionBoxesDatasetConfig,
)
from ..metadata.amazon import (
    AmazonBookItemMetadata,
    AmazonMovieItemMetadata,
    AmazonMusicItemMetadata,
    AmazonSubscriptionBoxesItemMetadata,
)
from .base import Dataset


logger = logging.getLogger(__name__)


class AmazonDataset(Dataset):
    """Base class for Amazon datasets.

    Other Amazon datasets should inherit from this class.
    """

    ITEM_METADATA = None
    IS_BASE: bool = True

    config: ClassVar[AmazonDatasetConfig] = AmazonDatasetConfig()

    def _download_dataset(self) -> None:
        """Downloads the dataset.

        Downloads the csv file from the dataset URL and saves it to the file path.
        """
        if not self.config.dataset_url:
            raise ValueError(f"{self.name} does not have URL specified in config.")

        logger.debug(f"Downloading {self.name} dataset from {self.config.dataset_url}")
        self._fetch_remote(
            self.config.dataset_url,
            self.file_path,
        )

    def _load_dataframe(self) -> pd.DataFrame:
        """Load the raw dataset from file, and return it as a pandas DataFrame.

        Transform the dataset downloaded to have integer user and item ids. This
        will be needed for representation in the interaction matrix.

        Returns:
            The interaction data as a DataFrame with a row per interaction.
        """
        self.fetch_dataset()

        # Read JSONL in chunks and show progress per chunk. We import tqdm
        # locally to avoid global pandas monkeypatching (`tqdm.pandas()`).
        from tqdm.auto import tqdm

        chunksize = 100_000
        chunks = pd.read_json(
            self.file_path,
            dtype={
                self.config.item_ix: str,
                self.config.user_ix: str,
                self.config.timestamp_ix: np.int64,
                self.config.rating_ix: np.float32,
                self.config.helpful_vote_ix: np.int64,
            },
            lines=True,
            chunksize=chunksize,
        )
        df = pd.concat(
            [chunk for chunk in tqdm(chunks, desc="Reading JSONL", unit="chunk")], ignore_index=True
        )

        df = df[
            [
                self.config.item_ix,
                self.config.user_ix,
                self.config.timestamp_ix,
                self.config.rating_ix,
                self.config.helpful_vote_ix,
            ]
        ]

        # Convert nanosecond timestamps to seconds
        df[self.config.timestamp_ix] = df[self.config.timestamp_ix] // 1_000_000_000

        logger.debug(f"Loaded {len(df)} interactions")
        return df


class AmazonMusicDataset(AmazonDataset):
    """Handles Amazon Music dataset."""

    IS_BASE: bool = False

    config: ClassVar[AmazonMusicDatasetConfig] = AmazonMusicDatasetConfig()

    def _fetch_dataset_metadata(
        self, user_id_mapping: pd.DataFrame, item_id_mapping: pd.DataFrame
    ) -> None:
        self.ITEM_METADATA = AmazonMusicItemMetadata(item_id_mapping=item_id_mapping).load()


class AmazonMovieDataset(AmazonDataset):
    """Handles Amazon Movie dataset."""

    IS_BASE: bool = False

    config: ClassVar[AmazonMovieDatasetConfig] = AmazonMovieDatasetConfig()

    def _fetch_dataset_metadata(
        self, user_id_mapping: pd.DataFrame, item_id_mapping: pd.DataFrame
    ) -> None:
        self.ITEM_METADATA = AmazonMovieItemMetadata(item_id_mapping=item_id_mapping).load()


class AmazonSubscriptionBoxesDataset(AmazonDataset):
    """Handles Amazon Computer dataset."""

    IS_BASE: bool = False

    config: ClassVar[AmazonSubscriptionBoxesDatasetConfig] = AmazonSubscriptionBoxesDatasetConfig()

    def _fetch_dataset_metadata(
        self, user_id_mapping: pd.DataFrame, item_id_mapping: pd.DataFrame
    ) -> None:
        self.ITEM_METADATA = AmazonSubscriptionBoxesItemMetadata(
            item_id_mapping=item_id_mapping
        ).load()


class AmazonBookDataset(AmazonDataset):
    """Handles Amazon Book dataset."""

    IS_BASE: bool = False

    config: ClassVar[AmazonBookDatasetConfig] = AmazonBookDatasetConfig()

    def _fetch_dataset_metadata(
        self, user_id_mapping: pd.DataFrame, item_id_mapping: pd.DataFrame
    ) -> None:
        self.ITEM_METADATA = AmazonBookItemMetadata(item_id_mapping=item_id_mapping).load()



import importlib
import logging
import uuid
from typing import Any, NamedTuple

from ..algorithms import Algorithm
from .base import Registry


logger = logging.getLogger(__name__)


class AlgorithmRegistry(Registry[Algorithm]):
    """Registry for easy retrieval of algorithm types by name.

    The registry is pre-registered with all recnexteval algorithms.
    """

    def __init__(self) -> None:
        """Initialize the algorithm registry.

        The registry is initialized with the `recnexteval.algorithms` module
        so that all built-in algorithms are available by default.
        """
        module = importlib.import_module("recnexteval.algorithms")
        super().__init__(module)


ALGORITHM_REGISTRY = AlgorithmRegistry()
"""Registry instantiation for algorithms.

Contains the recnexteval algorithms by default and allows registration of
new algorithms via the `register` function.

Examples:
    ```python
    from recnexteval.pipelines import ALGORITHM_REGISTRY

    # Construct an ItemKNN object with parameter K=20
    algo = ALGORITHM_REGISTRY.get("ItemKNN")(K=20)

    from recnexteval.algorithms import ItemKNN

    ALGORITHM_REGISTRY.register("HelloWorld", ItemKNN)

    # Also construct an ItemKNN object with parameter K=20
    algo = ALGORITHM_REGISTRY.get("HelloWorld")(K=20)
    ```
"""


class AlgorithmEntry(NamedTuple):
    """Entry for the algorithm registry.

    The intended use of this class is to store the name of the algorithm and
    the parameters that the algorithm should take. Mainly this is used during
    the building phase of the evaluator pipeline in `Builder`.

    Attributes:
        name: Name of the algorithm.
        params: Parameters that do not require optimization as key-value
            pairs, where the key is the hyperparameter name and the value is
            the value it should take.
    """

    name: str
    uuid: uuid.UUID | None = None
    params: None | dict[str, Any] = None



import logging
from warnings import warn

import numpy as np

from .top_k import MetricTopK


logger = logging.getLogger(__name__)


class ElementwiseMetricK(MetricTopK):
    """Base class for all elementwise metrics that can be calculated for
    each user-item pair in the Top-K recommendations.

    :attr:`results` contains an entry for each user-item pair.

    Examples are: HitK

    This code is adapted from RecPack :cite:`recpack`

    :param K: Size of the recommendation list consisting of the Top-K item predictions.
    :type K: int
    """

    # TODO to fix this function
    @property
    def micro_result(self) -> dict[str, np.ndarray]:
        if not self._is_computed:
            raise ValueError("Metric has not been calculated yet.")
        elif self._scores is None:
            warn(UserWarning("No scores were computed. Returning empty dict."))
            return dict(zip(self.col_names, (np.array([]), np.array([]))))

        scores = self._scores.toarray().reshape(-1)
        unique_users, inv = np.unique(self._user_id_sequence_array, return_inverse=True)

        # Sum hits per user
        sum_ones = np.zeros(len(unique_users))
        np.add.at(sum_ones, inv, scores)

        # Count recommendations per user
        count_all = np.zeros(len(unique_users))
        np.add.at(count_all, inv, 1)

        # aggregated score per user
        agg_score = sum_ones / count_all

        return dict(zip(self.col_names, (unique_users, agg_score)))


    @property
    def macro_result(self) -> None | float:
        if not self._is_computed:
            raise ValueError("Metric has not been calculated yet.")
        elif self._scores is None:
            logger.warning(UserWarning("No scores were computed. Returning Null value."))
            return None
        elif self._scores.size == 0:
            logger.warning(
                UserWarning(
                    f"All predictions were off or the ground truth matrix was empty during compute of {self.identifier}."
                )
            )
            return 0

        scores = self._scores.toarray().reshape(-1)
        unique_users, inv = np.unique(self._user_id_sequence_array, return_inverse=True)
        # get all users that was recommended at least a relevant item
        sum_ones = np.zeros(len(unique_users))
        np.add.at(sum_ones, inv, scores)
        # Convert to binary: 1 if at least 1 hit, 0 otherwise
        binary_hits = (sum_ones > 0).astype(int)
        # Fraction of users with at least 1 hit
        return binary_hits.mean().item()



# RecPack, An Experimentation Toolkit for Top-N Recommendation
# Copyright (C) 2020  Froomle N.V.
# License: GNU AGPLv3 - https://gitlab.com/recpack-maintainers/recpack/-/blob/master/LICENSE
# Author:
#   Lien Michiels
#   Robin Verachtert

import logging

import numpy as np
from scipy.sparse import csr_matrix


EPSILON = 1e-13

logger = logging.getLogger(__name__)


def invert(x: np.ndarray | csr_matrix) -> np.ndarray | csr_matrix:
    """Invert an array.

    :param x: [description]
    :type x: [type]
    :return: [description]
    :rtype: [type]
    """
    if isinstance(x, np.ndarray):
        ret = np.zeros(x.shape)
    elif isinstance(x, csr_matrix):
        ret = csr_matrix(x.shape)
    else:
        raise TypeError("Unsupported type for argument x.")
    ret[x.nonzero()] = 1 / x[x.nonzero()]
    return ret


def to_binary(X: csr_matrix) -> csr_matrix:
    """Converts a matrix to binary by setting all non-zero values to 1.

    :param X: Matrix to convert to binary.
    :type X: csr_matrix
    :return: Binary matrix.
    :rtype: csr_matrix
    """
    X_binary = X.astype(bool).astype(X.dtype)

    return X_binary


import logging
import os
import zipfile
from typing import ClassVar

import numpy as np
import pandas as pd
from tqdm import tqdm

from ..config import MovieLens100KDatasetConfig, MovieLensDatasetConfig
from ..metadata.movielens import MovieLens100kItemMetadata, MovieLens100kUserMetadata
from .base import Dataset


logger = logging.getLogger(__name__)


class MovieLensDataset(Dataset):
    """Base class for Movielens datasets.

    Other Movielens datasets should inherit from this class.

    This code is adapted from RecPack :cite:`recpack`
    """

    IS_BASE: bool = True
    config: ClassVar[MovieLensDatasetConfig] = MovieLensDatasetConfig()

    def _download_dataset(self) -> None:
        # Download the zip into the data directory
        zip_file_path = os.path.join(
            self.config.default_base_path, f"{self.config.remote_zipname}.zip"
        )
        self._fetch_remote(
            url=f"{self.config.dataset_url}/{self.config.remote_zipname}.zip",
            filename=zip_file_path,
        )

        # Extract the ratings file which we will use
        with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
            zip_ref.extract(
                f"{self.config.remote_zipname}/{self.config.remote_filename}",
                self.config.default_base_path,
            )

        # Rename the ratings file to the specified filename
        os.rename(
            os.path.join(
                self.config.default_base_path,
                f"{self.config.remote_zipname}/{self.config.remote_filename}",
            ),
            self.file_path,
        )


class MovieLens100K(MovieLensDataset):
    """MovieLens 100K dataset."""

    ITEM_METADATA = None
    USER_METADATA = None
    IS_BASE: bool = False

    config: ClassVar[MovieLens100KDatasetConfig] = MovieLens100KDatasetConfig()

    def _load_dataframe(self) -> pd.DataFrame:
        self.fetch_dataset()
        chunks = pd.read_table(
            self.file_path,
            dtype={
                self.config.user_ix: np.int64,
                self.config.item_ix: np.int64,
                self.config.rating_ix: np.float64,
                self.config.timestamp_ix: np.int64,
            },
            sep="\t",
            names=[
                self.config.user_ix,
                self.config.item_ix,
                self.config.rating_ix,
                self.config.timestamp_ix,
            ],
            chunksize=100_000,
        )
        df = pd.concat(
            [chunk for chunk in tqdm(chunks, desc="Reading table", unit="chunk")], ignore_index=True
        )
        return df

    def _fetch_dataset_metadata(
        self, user_id_mapping: pd.DataFrame, item_id_mapping: pd.DataFrame
    ) -> None:
        self.USER_METADATA = MovieLens100kUserMetadata(user_id_mapping=user_id_mapping).load()
        self.ITEM_METADATA = MovieLens100kItemMetadata(item_id_mapping=item_id_mapping).load()



import logging
from collections.abc import Iterator
from dataclasses import dataclass, field
from enum import StrEnum
from typing import Any
from uuid import UUID

from recnexteval.algorithms import Algorithm
from ..utils.uuid_util import generate_algorithm_uuid


logger = logging.getLogger(__name__)


class AlgorithmStateEnum(StrEnum):
    """Enum for the state of the algorithm.

    Used to keep track of the state of the algorithm during the streaming
    process in the `EvaluatorStreamer`.
    """

    NEW = "NEW"
    READY = "READY"
    RUNNING = "RUNNING"
    PREDICTED = "PREDICTED"
    COMPLETED = "COMPLETED"


@dataclass
class AlgorithmStateEntry:
    """Entry for the algorithm status registry.

    This dataclass stores the status of an algorithm for use by
    `AlgorithmStateManager`. It contains the algorithm name, unique
    identifier, current state, associated data segment, and an optional
    pointer to the algorithm object.

    Attributes:
        name: Name of the algorithm.
        algo_uuid: Unique identifier for the algorithm.
        state: State of the algorithm.
        data_segment: Data segment the algorithm is associated with.
        params: Parameters for the algorithm.
        algo_ptr: Pointer to the algorithm object.
    """

    name: str
    algorithm_uuid: UUID
    state: AlgorithmStateEnum = AlgorithmStateEnum.NEW
    data_segment: int = 0
    params: dict[str, Any] = field(default_factory=dict)
    algo_ptr: None | type[Algorithm] | Algorithm = None


class AlgorithmStateManager:
    def __init__(self) -> None:
        self._algorithms: dict[UUID, AlgorithmStateEntry] = {}

    def __iter__(self) -> Iterator[UUID]:
        """Return an iterator over registered algorithm UUIDs.

        Allows iteration over the UUIDs of registered entries.

        Returns:
            An iterator over the UUIDs of registered entries.
        """
        return iter(self._algorithms)

    def __len__(self) -> int:
        """Return the number of registered algorithms.

        Returns:
            The number of registered algorithms.
        """
        return len(self._algorithms)

    def values(self) -> Iterator[AlgorithmStateEntry]:
        """Return an iterator over registered AlgorithmStateEntry objects.

        Allows iteration over the registered entries.

        Returns:
            An iterator over the registered entries.
        """
        return iter(self._algorithms.values())

    def __getitem__(self, key: UUID) -> AlgorithmStateEntry:
        if key not in self._algorithms:
            raise ValueError(f"Algorithm with ID:{key} not registered")
        return self._algorithms[key]

    def __setitem__(self, key: UUID, entry: AlgorithmStateEntry) -> None:
        """Register a new algorithm status entry under `key`.

        Allows the use of square bracket notation to register new entries.

        Args:
            key: The UUID to register the entry under.
            entry: The status entry to register.

        Raises:
            KeyError: If `key` is already registered.
        """
        if key in self:
            raise KeyError(f"Algorithm with ID:{key} already registered")
        self._algorithms[key] = entry

    def __contains__(self, key: UUID) -> bool:
        """Return whether the given key is known to the registry.

        Args:
            key: The key to check.

        Returns:
            True if the key is registered, False otherwise.
        """
        try:
            self[key]
            return True
        except AttributeError:
            return False

    def get(self, algo_id: UUID) -> AlgorithmStateEntry:
        """Get the :class:`AlgorithmStateEntry` for `algo_id`."""
        return self[algo_id]

    def get_state(self, algo_id: UUID) -> AlgorithmStateEnum:
        """Get the current state of the algorithm with `algo_id`."""
        return self[algo_id].state

    def register(
        self,
        name: None | str = None,
        algo_ptr: None | type[Algorithm] | Algorithm = None,
        params: dict[str, Any] = {},
        algo_uuid: None | UUID = None,
    ) -> UUID:
        """Register new algorithm"""
        if not name and not algo_ptr:
            raise ValueError("Either name or algo_ptr must be provided for registration")
        elif algo_ptr and isinstance(algo_ptr, type):
            algo_ptr = algo_ptr(**params)
            name = name or algo_ptr.identifier
        elif algo_ptr and hasattr(algo_ptr, "identifier") and not name:
            name = name or algo_ptr.identifier  # type: ignore[attr-defined]
        elif not name:
            # This should not happen if name was provided or algo_ptr has identifier
            raise ValueError("Algorithm name was not provided and could not be inferred from Algorithm pointer")

        if algo_uuid is None:
            algo_uuid = generate_algorithm_uuid(name)

        entry = AlgorithmStateEntry(algorithm_uuid=algo_uuid, name=name, algo_ptr=algo_ptr, params=params)
        self._algorithms[algo_uuid] = entry
        logger.info(f"Registered algorithm '{name}' with ID {algo_uuid}")
        return algo_uuid

    def can_request_training_data(self, algo_id: UUID) -> tuple[bool, str]:
        """Check if algorithm can request training data"""
        if algo_id not in self._algorithms:
            return False, f"Algorithm {algo_id} not registered"

        state = self._algorithms[algo_id].state

        if state == AlgorithmStateEnum.COMPLETED:
            return False, "Algorithm has completed evaluation"
        if state == AlgorithmStateEnum.NEW:
            return False, "The algorithm must be set to READY state first"
        if state == AlgorithmStateEnum.PREDICTED:
            return False, "Algorithm has already requested data for this window"
        if state == AlgorithmStateEnum.READY:
            return True, ""

        return False, f"Unknown state {state}"

    def can_request_unlabeled_data(self, algo_id: UUID) -> tuple[bool, str]:
        """Check if algorithm can request unlabeled data"""
        if algo_id not in self._algorithms:
            return False, f"Algorithm {algo_id} not registered"

        state = self._algorithms[algo_id].state

        if state == AlgorithmStateEnum.RUNNING:
            return True, ""
        if state == AlgorithmStateEnum.COMPLETED:
            return False, "Algorithm has completed evaluation"
        if state == AlgorithmStateEnum.NEW:
            return False, "The algorithm must be set to RUNNING state to request unlabeled data"
        if state == AlgorithmStateEnum.PREDICTED:
            return False, "Algorithm has already requested data for this window"
        if state == AlgorithmStateEnum.READY:
            return (
                False,
                "The algorithm must be set to RUNNING state to request unlabeled data. Request training data first",
            )

        return False, f"Unknown state {state}"

    def can_submit_prediction(self, algo_id: UUID) -> tuple[bool, str]:
        """Check if algorithm can submit prediction"""
        if algo_id not in self._algorithms:
            return False, f"Algorithm {algo_id} not registered"

        state = self._algorithms[algo_id].state

        if state == AlgorithmStateEnum.RUNNING:
            return True, ""
        if state == AlgorithmStateEnum.READY:
            return False, "There is new data to be requested"
        if state == AlgorithmStateEnum.NEW:
            return False, "Algorithm must request data first"
        if state == AlgorithmStateEnum.PREDICTED:
            return False, "Algorithm already submitted prediction for this window"
        if state == AlgorithmStateEnum.COMPLETED:
            return False, "Algorithm has completed evaluation"

        return False, f"Unknown state {state}"

    def transition(self, algo_id: UUID, new_state: AlgorithmStateEnum, data_segment: None | int = None) -> None:
        """Transition algorithm to new state with validation"""
        if algo_id not in self._algorithms:
            raise ValueError(f"Algorithm {algo_id} not registered")

        entry = self._algorithms[algo_id]
        old_state = entry.state

        # Define valid transitions
        valid_transitions = {
            # old_state: [list of valid new_states]
            AlgorithmStateEnum.NEW: [AlgorithmStateEnum.READY, AlgorithmStateEnum.COMPLETED],
            AlgorithmStateEnum.READY: [AlgorithmStateEnum.RUNNING],
            AlgorithmStateEnum.RUNNING: [AlgorithmStateEnum.PREDICTED],
            AlgorithmStateEnum.PREDICTED: [AlgorithmStateEnum.READY, AlgorithmStateEnum.COMPLETED],
            AlgorithmStateEnum.COMPLETED: [],
        }

        if new_state not in valid_transitions.get(old_state, []):
            raise ValueError(f"Invalid transition: {old_state} -> {new_state}")

        entry.state = new_state
        if data_segment is not None:
            entry.data_segment = data_segment

        logger.debug(f"Algorithm '{entry.name}' transitioned {old_state.value} -> {new_state.value}")

    def is_all_predicted(self) -> bool:
        """Return whether every registered algorithm is in PREDICTED state.

        Returns:
            True if all registered entries have state
            `AlgorithmStateEnum.PREDICTED`, False otherwise.
        """
        if not self._algorithms:
            return False
        return all(entry.state == AlgorithmStateEnum.PREDICTED for entry in self._algorithms.values())

    def get_all_states(self) -> dict[str, AlgorithmStateEnum]:
        """Get state of all algorithms"""
        return {entry.name: entry.state for entry in self._algorithms.values()}

    def is_all_same_data_segment(self) -> bool:
        """Return whether all registered entries share the same data segment.

        Returns:
            True if there is exactly one distinct data segment across all
            registered entries, False otherwise.
        """
        data_segments: set[None | int] = set()
        for key in self:
            data_segments.add(self[key].data_segment)
        return len(data_segments) == 1

    def all_algo_states(self) -> dict[str, AlgorithmStateEnum]:
        """Return a mapping of identifier strings to algorithm states.

        The identifier used is "{name}_{uuid}" for each registered entry.

        Returns:
            Mapping from identifier string to the entry's
            :class:`AlgorithmStateEnum`.
        """
        states: dict[str, AlgorithmStateEnum] = {}
        for key in self:
            states[f"{self[key].name}_{key}"] = self[key].state
        return states

    def set_all_ready(self, data_segment: int) -> None:
        """Set all registered algorithms to the READY state.

        Args:
            data_segment: Data segment to assign to every algorithm.
        """
        for key in self:
            self.transition(key, AlgorithmStateEnum.READY, data_segment)

    def get_algorithm_identifier(self, algo_id: UUID) -> str:
        """Return a stable identifier string for the algorithm.

        Args:
            algo_id: UUID of the algorithm.

        Returns:
            Identifier in the format "{name}_{uuid}".

        Raises:
            AttributeError: If `algo_id` is not registered.
        """
        if algo_id not in self._algorithms:
            raise AttributeError(f"Algorithm with ID:{algo_id} not registered")
        return f"{self[algo_id].name}_{algo_id}"



import os
import zipfile
from typing import ClassVar

import pandas as pd

from recnexteval.datasets.config import (
    MovieLens100kItemMetadataConfig,
    MovieLens100kUserMetadataConfig,
)
from .base import Metadata


class MovieLens100kMetadata(Metadata):
    def _download_dataset(self) -> None:
        # Download the zip into the data directory
        self._fetch_remote(
            f"{self.config.dataset_url}/{self.config.remote_zipname}.zip",
            os.path.join(self.base_path, f"{self.config.remote_zipname}.zip"),
        )

        # Extract the ratings file which we will use
        with zipfile.ZipFile(
            os.path.join(self.base_path, f"{self.config.remote_zipname}.zip"), "r"
        ) as zip_ref:
            zip_ref.extract(
                f"{self.config.remote_zipname}/{self.config.remote_filename}", self.base_path
            )

        # Rename the ratings file to the specified filename
        os.rename(
            os.path.join(
                self.base_path, f"{self.config.remote_zipname}/{self.config.remote_filename}"
            ),
            self.file_path,
        )


class MovieLens100kUserMetadata(MovieLens100kMetadata):
    config: ClassVar[MovieLens100kUserMetadataConfig] = MovieLens100kUserMetadataConfig()  # type: ignore

    def __init__(self, user_id_mapping: pd.DataFrame) -> None:
        super().__init__()
        self.user_id_mapping = user_id_mapping

    def _load_dataframe(self) -> pd.DataFrame:
        self.fetch_dataset()
        df = pd.read_table(
            self.file_path,
            dtype=self.config.dtype_dict,
            sep=self.config.sep,
            names=self.config.column_names,
            converters={self.config.user_ix: self._map_user_id},
        )
        return df

    def _map_user_id(self, user_id):
        user_id_to_uid = dict(
            zip(self.user_id_mapping[self.config.user_ix], self.user_id_mapping["uid"])
        )
        return user_id_to_uid.get(int(user_id), user_id)


class MovieLens100kItemMetadata(MovieLens100kMetadata):
    config: ClassVar[MovieLens100kItemMetadataConfig] = MovieLens100kItemMetadataConfig()  # type: ignore

    def __init__(self, item_id_mapping: pd.DataFrame) -> None:
        super().__init__()
        self.item_id_mapping = item_id_mapping

    def _load_dataframe(self) -> pd.DataFrame:
        self.fetch_dataset()
        df = pd.read_table(
            self.file_path,
            dtype=self.config.dtype_dict,
            sep=self.config.sep,
            names=self.config.column_names,
            converters={self.config.item_ix: self._map_item_id},
            encoding=self.config.encoding,
        )
        return df

    def _map_item_id(self, item_id):
        item_id_to_iid = dict(
            zip(self.item_id_mapping[self.config.item_ix], self.item_id_mapping["iid"])
        )
        return item_id_to_iid.get(int(item_id), item_id)



from dataclasses import dataclass

from .base import DatasetConfig, MetadataConfig


@dataclass
class AmazonDatasetConfig(DatasetConfig):
    """Amazon dataset base configuration."""

    user_ix: str = "user_id"
    item_ix: str = "parent_asin"
    timestamp_ix: str = "timestamp"
    rating_ix: str = "rating"
    helpful_vote_ix: str = "helpful_vote"
    dataset_url: str = ""  # Set per subclass
    remote_filename: str = ""  # Set per subclass
    default_base_path: str = DatasetConfig.default_base_path + "/amazon"

    @property
    def default_filename(self) -> str:
        """Return just the filename for Amazon datasets (no zipname prefix)."""
        return self.remote_filename


@dataclass
class AmazonMusicDatasetConfig(AmazonDatasetConfig):
    """Amazon Music dataset configuration."""

    remote_filename: str = "Digital_Music.jsonl.gz"
    dataset_url: str = "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/raw/review_categories/Digital_Music.jsonl.gz"


@dataclass
class AmazonMovieDatasetConfig(AmazonDatasetConfig):
    """Amazon Movie dataset configuration."""

    remote_filename: str = "Movies_and_TV.jsonl.gz"
    dataset_url: str = "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/raw/review_categories/Movies_and_TV.jsonl.gz"


@dataclass
class AmazonSubscriptionBoxesDatasetConfig(AmazonDatasetConfig):
    """Amazon Subscription Boxes dataset configuration."""

    remote_filename: str = "Subscription_Boxes.jsonl.gz"
    dataset_url: str = "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/raw/review_categories/Subscription_Boxes.jsonl.gz"


@dataclass
class AmazonBookDatasetConfig(AmazonDatasetConfig):
    """Amazon Books dataset configuration."""

    remote_filename: str = "Books.jsonl.gz"
    dataset_url: str = "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/raw/review_categories/Books.jsonl.gz"


@dataclass
class AmazonItemMetadataConfig(MetadataConfig, AmazonDatasetConfig):
    """
    Amazon Item Metadata Base Configuration.

    Handles configuration for Amazon product metadata including:
    - Product identifiers (ASIN)
    - Product information (title, category, price, rating)
    - Rich content (features, description, images, videos)
    - Relational data (store, details, bought together)

    All properties are computed from base fields to ensure consistency.
    """

    item_ix: str = "parent_asin"
    """Name of the column containing product identifiers (parent ASIN)."""
    main_category_ix: str = "main_category"
    """Name of the column containing the main product category."""
    title_ix: str = "title"
    """Name of the column containing product title."""
    average_rating_ix: str = "average_rating"
    """Name of the column containing average product rating (0-5)."""
    rating_number_ix: str = "rating_number"
    """Name of the column containing number of ratings received."""
    features_ix: str = "features"
    """Name of the column containing product features (list)."""
    description_ix: str = "description"
    """Name of the column containing product description (list)."""
    price_ix: str = "price"
    """Name of the column containing product price."""
    images_ix: str = "images"
    """Name of the column containing product images URLs (list)."""
    videos_ix: str = "videos"
    """Name of the column containing product videos URLs (list)."""
    store_ix: str = "store"
    """Name of the column containing store/seller information."""
    categories_ix: str = "categories"
    """Name of the column containing category hierarchy (list)."""
    details_ix: str = "details"
    """Name of the column containing product details (dict)."""
    bought_together_ix: str = "bought_together"
    """Name of the column containing products bought together (list)."""

    @property
    def column_names(self) -> list[str]:
        return [
            self.main_category_ix,
            self.title_ix,
            self.average_rating_ix,
            self.rating_number_ix,
            self.features_ix,
            self.description_ix,
            self.price_ix,
            self.images_ix,
            self.videos_ix,
            self.store_ix,
            self.categories_ix,
            self.details_ix,
            self.item_ix,
            self.bought_together_ix,
        ]

    @property
    def dtype_dict(self) -> dict:
        return {
            self.main_category_ix: str,
            self.title_ix: str,
            self.average_rating_ix: "float32",
            self.rating_number_ix: "int64",
            self.features_ix: list,
            self.description_ix: list,
            self.price_ix: "float32",
            self.images_ix: list,
            self.videos_ix: list,
            self.store_ix: str,
            self.categories_ix: list,
            self.details_ix: dict,
            self.item_ix: str,
            self.bought_together_ix: list,
        }


@dataclass
class AmazonDigitalMusicItemMetadataConfig(AmazonItemMetadataConfig):
    """Amazon Digital Music metadata configuration."""

    remote_filename: str = "meta_Digital_Music.jsonl.gz"
    """Filename for Digital Music metadata."""

    dataset_url: str = (
        "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/"
        "raw/meta_categories/meta_Digital_Music.jsonl.gz"
    )


@dataclass
class AmazonMoviesAndTVItemMetadataConfig(AmazonItemMetadataConfig):
    """Amazon Movies and TV metadata configuration."""

    remote_filename: str = "meta_Movies_and_TV.jsonl.gz"
    """Filename for Movies and TV metadata."""

    dataset_url: str = (
        "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/"
        "raw/meta_categories/meta_Movies_and_TV.jsonl.gz"
    )


@dataclass
class AmazonSubscriptionBoxesItemMetadataConfig(AmazonItemMetadataConfig):
    """Amazon Subscription Boxes metadata configuration."""

    remote_filename: str = "meta_Subscription_Boxes.jsonl.gz"
    """Filename for Subscription Boxes metadata."""

    dataset_url: str = (
        "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/"
        "raw/meta_categories/meta_Subscription_Boxes.jsonl.gz"
    )


@dataclass
class AmazonBooksItemMetadataConfig(AmazonItemMetadataConfig):
    """Amazon Books metadata configuration."""

    remote_filename: str = "meta_Books.jsonl.gz"
    """Filename for Books metadata."""

    dataset_url: str = (
        "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/"
        "raw/meta_categories/meta_Books.jsonl.gz"
    )



# Adopted from RecPack, An Experimentation Toolkit for Top-N Recommendation
# Copyright (C) 2020  Froomle N.V.
# License: GNU AGPLv3 - https://gitlab.com/recpack-maintainers/recpack/-/blob/master/LICENSE
# Author:
#   Lien Michiels
#   Robin Verachtert

from recnexteval.algorithms.time_aware_item_knn.base import TARSItemKNN


class TARSItemKNNDing(TARSItemKNN):
    """Time aware variant of ItemKNN which uses an exponential decay function at prediction time and cosine similarity.

    Algorithm as presented in
    Yi Ding and Xue Li. 2005.
    Time weight collaborative filtering.
    In Proceedings of the 14th ACM international conference on Information and knowledge management (CIKM '05).
    Association for Computing Machinery, New York, NY, USA, 485–492.
    https://doi.org/10.1145/1099554.1099689


    Computation of the similarity matrix is the same as normal ItemKNN.
    When predicting however the user's older interactions are given less weight in the final prediction score.

    .. math::

        \\text{sim}(u, i) = \\sum\\limits_{j \\in X_u} e^{-\\alpha \\cdot \\delta t_{u,j}} \\cdot \\text{sim}(i, j)

    Where :math:`\\alpha` is the `predict_decay` parameter.

    :param K: How many neigbours to use per item,
        make sure to pick a value below the number of columns of the matrix to fit on.
        Defaults to 200
    :type K: int, Optional
    :param pad_with_popularity: Whether to pad the similarity matrix with RecentPop Algorithm.
        Defaults to True.
    :type pad_with_popularity: bool, optional
    :param predict_decay: Defines the decay scaling used for decay during prediction.
        Defaults to 1 / (24 * 3600).
        This means for every day since an interaction, the value of it will be divided by 'e'.
    :type predict_decay: float, optional
    :param similarity: Which similarity measure to use. Defaults to `"cosine"`.
        ``["cosine", "conditional_probability"]`` are supported.
    :type similarity: str, optional

    This code is adapted from RecPack :cite:`recpack`
    """

    SUPPORTED_SIMILARITIES = ["cosine", "conditional_probability"]

    def __init__(self, K: int = 200, pad_with_popularity: bool = True, predict_decay: float = 1 / (24 * 3600), similarity: str = "cosine"):
        super().__init__(K=K, pad_with_popularity=pad_with_popularity, fit_decay=0, predict_decay=predict_decay, similarity=similarity, decay_function="exponential")



"""Registries for algorithms, metrics, and datasets.

This module provides registries for storing and managing algorithms, metrics,
and datasets used in experiments. Registries help keep track of valid classes
and enable easy instantiation of components.

## Registries

Registries store algorithms, metrics, and datasets by default and allow
registration of new components via the `register` function.

Example:
    ```python
    from recnexteval.pipelines import ALGORITHM_REGISTRY
    from recnexteval.algorithms import ItemKNNStatic

    algo = ALGORITHM_REGISTRY.get("ItemKNNStatic")(K=10)
    ALGORITHM_REGISTRY.register("algo_1", ItemKNNStatic)
    ```

### Available Registries

- `ALGORITHM_REGISTRY`: Registry for algorithms
- `DATASET_REGISTRY`: Registry for datasets
- `METRIC_REGISTRY`: Registry for metrics
- `AlgorithmRegistry`: Class for creating algorithm registries
- `DatasetRegistry`: Class for creating dataset registries
- `MetricRegistry`: Class for creating metric registries

## Entries

Entries store algorithms and metrics in registries. They maintain the class
and parameters used to instantiate each component. These entries are used by
`EvaluatorPipeline` to instantiate algorithms and metrics.

### Available Entries

- `AlgorithmEntry`: Entry for algorithms
- `MetricEntry`: Entry for metrics

"""

from .algorithm import (
    ALGORITHM_REGISTRY,
    AlgorithmEntry,
    AlgorithmRegistry,
)
from .base import Registry
from .dataset import DATASET_REGISTRY, DatasetRegistry
from .metric import (
    METRIC_REGISTRY,
    MetricEntry,
    MetricRegistry,
)


__all__ = [
    "ALGORITHM_REGISTRY",
    "AlgorithmEntry",
    "AlgorithmRegistry",
    "DATASET_REGISTRY",
    "DatasetRegistry",
    "METRIC_REGISTRY",
    "MetricEntry",
    "MetricRegistry",
    "Registry",
]



import logging
import uuid
from warnings import warn

from recnexteval.algorithms import Algorithm
from recnexteval.evaluators.evaluator_pipeline import EvaluatorPipeline
from ..state_management import AlgorithmStateManager
from .base import Builder


logger = logging.getLogger(__name__)


class EvaluatorPipelineBuilder(Builder):
    """Builder to facilitate construction of evaluator.

    Provides methods to set specific values for the evaluator and enforce checks
    such that the evaluator can be constructed correctly and to avoid possible
    errors when the evaluator is executed.
    """

    def __init__(
        self,
        ignore_unknown_user: bool = False,
        ignore_unknown_item: bool = False,
        seed: int = 42,
    ) -> None:
        """Initialize the EvaluatorPipelineBuilder.

        Args:
            ignore_unknown_user: Ignore unknown user in the evaluation.
            ignore_unknown_item: Ignore unknown item in the evaluation.
            seed: Random seed for reproducibility.
        """
        super().__init__(
            ignore_unknown_user=ignore_unknown_user,
            ignore_unknown_item=ignore_unknown_item,
            seed=seed,
        )
        self.algo_state_mgr = AlgorithmStateManager()

    def add_algorithm(
        self,
        algorithm: type[Algorithm],
        params: dict[str, int] = {},
        algo_uuid: None | uuid.UUID = None,
    ) -> None:
        """Add algorithm to evaluate.

        Adding algorithm to evaluate on. The algorithm can be added by specifying the class type
        or by specifying the class name as a string.

        Args:
            algorithm: Algorithm to evaluate.
            params: Parameter for the algorithm.

        Raises:
            ValueError: If algorithm is not found in ALGORITHM_REGISTRY.
        """
        if not self._check_setting_exist():
            raise RuntimeError(
                "Setting has not been set. To ensure conformity, of the addition of"
                " other components please set the setting first. Call add_setting() method."
            )

        self.algo_state_mgr.register(algo_ptr=algorithm, params=params, algo_uuid=algo_uuid)

    def _check_ready(self) -> None:
        """Check if the builder is ready to construct Evaluator.

        Raises:
            RuntimeError: If there are invalid configurations.
        """
        super()._check_ready()

        if len(self.algo_state_mgr) == 0:
            raise RuntimeError("No algorithms specified, can't construct Evaluator")

        for algo_state in self.algo_state_mgr.values():
            if (
                algo_state.params is not None
                and "K" in algo_state.params
                and algo_state.params["K"] < self.setting.top_K
            ):
                warn(
                    f"Algorithm {algo_state.name} has K={algo_state.params['K']} but setting"
                    f" is configured top_K={self.setting.top_K}. The number of predictions"
                    " returned by the model is less than the K value to evaluate the predictions"
                    " on. This may distort the metric value."
                )

    def build(self) -> EvaluatorPipeline:
        """Build Evaluator object.

        Raises:
            RuntimeError: If no metrics, algorithms or settings are specified.

        Returns:
            EvaluatorPipeline: The built evaluator object.
        """
        self._check_ready()
        return EvaluatorPipeline(
            # algorithm_entries=self.algorithm_entries,
            algo_state_mgr=self.algo_state_mgr,
            metric_entries=list(self.metric_entries.values()),
            setting=self.setting,
            metric_k=self.metric_k,
            ignore_unknown_user=self.ignore_unknown_user,
            ignore_unknown_item=self.ignore_unknown_item,
            seed=self.seed,
        )



import logging

import scipy.sparse
from scipy.sparse import csr_matrix

from ..core.listwise_top_k import ListwiseMetricK


logger = logging.getLogger(__name__)


class RecallK(ListwiseMetricK):
    """Computes the fraction of true interactions that made it into
    the Top-K recommendations.

    Recall per user is computed as:

    .. math::

        \\text{Recall}(u) = \\frac{\\sum\\limits_{i \\in \\text{Top-K}(u)} y^{true}_{u,i} }{\\sum\\limits_{j \\in I} y^{true}_{u,j}}

    ref: RecPack

    :param K: Size of the recommendation list consisting of the Top-K item predictions.
    :type K: int
    """
    IS_BASE: bool = False

    def _calculate(self, y_true: csr_matrix, y_pred: csr_matrix) -> None:
        scores = scipy.sparse.lil_matrix(y_pred.shape)

        logger.debug("Precision compute started - %s", self.name)
        logger.debug("Shape of matrix: (%d, %d)", y_true.shape[0], y_true.shape[1])
        logger.debug("Number of ground truth interactions: %d", y_true.nnz)

        # obtain true positives
        scores[y_pred.multiply(y_true).astype(bool)] = 1
        scores = scores.tocsr()

        # true positive/total actual interactions
        self._scores = csr_matrix(scores.sum(axis=1) / y_true.sum(axis=1))
        logger.debug(f"Recall compute complete - {self.name}")



import logging
from collections import defaultdict
from typing import Optional

import pandas as pd

from recnexteval.metrics import Metric
from .util import MetricLevelEnum


logger = logging.getLogger(__name__)


class MetricAccumulator:
    def __init__(self) -> None:
        self.acc: defaultdict[str, dict[str, Metric]] = defaultdict(dict)

    def __getitem__(self, key) -> dict[str, Metric]:
        return self.acc[key]

    def add(self, metric: Metric, algorithm_name: str) -> None:
        """Add a metric to the accumulator

        Takes a :class:`Metric` object and adds it under the algorithm name. If
        the specified metric already exists for the algorithm, it will be
        overwritten with the new metric.

        :param metric: Metric to store
        :type metric: Metric
        :param algorithm_name: Name of the algorithm
        :type algorithm_name: str
        """
        if metric.identifier in self.acc[algorithm_name]:
            logger.warning(
                f"Metric {metric.identifier} already exists for algorithm {algorithm_name}. Overwriting..."
            )

        logger.debug(f"Metric {metric.identifier} created for algorithm {algorithm_name}")

        self.acc[algorithm_name][metric.identifier] = metric

    @property
    def user_level_metrics(self) -> defaultdict:
        results = defaultdict()
        for algo_name in self.acc:
            for metric_identifier in self.acc[algo_name]:
                metric = self.acc[algo_name][metric_identifier]
                results[(algo_name, f"t={metric.timestamp_limit}", metric.name)] = (
                    metric.micro_result
                )
        return results

    @property
    def window_level_metrics(self) -> defaultdict:
        results = defaultdict(dict)
        for algo_name in self.acc:
            for metric_identifier in self.acc[algo_name]:
                metric = self.acc[algo_name][metric_identifier]
                score = metric.macro_result
                num_user = metric.num_users
                if score == 0 and num_user == 0:
                    logger.info(
                        f"Metric {metric.name} for algorithm {algo_name} "
                        f"at t={metric.timestamp_limit} has 0 score and 0 users. "
                        "The ground truth may be empty due to no interactions occurring in that window."
                    )
                elif score == 0 and num_user != 0:
                    logger.info(
                        f"Metric {metric.name} for algorithm {algo_name} "
                        f"at t={metric.timestamp_limit} has 0 score but there are interactions. "
                        f"{algo_name} did not have any correct predictions."
                    )
                results[(algo_name, f"t={metric.timestamp_limit}", metric.name)]["score"] = score
                results[(algo_name, f"t={metric.timestamp_limit}", metric.name)]["num_user"] = (
                    num_user
                )
        return results

    def df_user_level_metric(self) -> pd.DataFrame:
        """User metric across all timestamps

        Computation of metrics evaluated on the user level

        :return: _description_
        :rtype: pd.DataFrame
        """
        df = pd.DataFrame.from_dict(self.user_level_metrics, orient="index").explode(
            ["user_id", "score"]
        )
        df = df.rename_axis(["algorithm", "timestamp", "metric"])
        df.rename(columns={"score": "user_score"}, inplace=True)
        return df

    def df_window_level_metric(self) -> pd.DataFrame:
        df = pd.DataFrame.from_dict(self.window_level_metrics, orient="index").explode(
            ["score", "num_user"]
        )
        df = df.rename_axis(["algorithm", "timestamp", "metric"])
        df.rename(columns={"score": "window_score"}, inplace=True)
        return df

    def df_macro_level_metric(self) -> pd.DataFrame:
        """Macro metric across all timestamps

        :return: _description_
        :rtype: pd.DataFrame
        """
        df = pd.DataFrame.from_dict(self.window_level_metrics, orient="index").explode(
            ["score", "num_user"]
        )
        df = df.rename_axis(["algorithm", "timestamp", "metric"])
        result = df.groupby(["algorithm", "metric"]).mean()["score"].to_frame()
        result["num_window"] = df.groupby(["algorithm", "metric"]).count()["score"]
        result = result.rename(columns={"score": "macro_score"})
        return result

    def df_micro_level_metric(self) -> pd.DataFrame:
        """Micro metric across all timestamps

        :return: _description_
        :rtype: pd.DataFrame
        """
        df = pd.DataFrame.from_dict(self.user_level_metrics, orient="index").explode(
            ["user_id", "score"]
        )
        df = df.rename_axis(["algorithm", "timestamp", "metric"])
        result = df.groupby(["algorithm", "metric"])["score"].mean().to_frame()
        result["num_user"] = df.groupby(["algorithm", "metric"])["score"].count()
        result = result.rename(columns={"score": "micro_score"})
        return result

    def df_metric(
        self,
        filter_timestamp: Optional[int] = None,
        filter_algo: Optional[str] = None,
        level: MetricLevelEnum = MetricLevelEnum.MACRO,
    ) -> pd.DataFrame:
        """Dataframe representation of the metric

        Returns a dataframe representation of the metric. The dataframe can be
        filtered based on the algorithm name and the timestamp.

        :param filter_timestamp: Timestamp value to filter on, defaults to None
        :type filter_timestamp: Optional[int], optional
        :param filter_algo: Algorithm name to filter on, defaults to None
        :type filter_algo: Optional[str], optional
        :param level: Level of the metric to compute, defaults to MetricLevelEnum.MACRO
        :type level: MetricLevelEnum, optional
        :return: Dataframe representation of the metric
        :rtype: pd.DataFrame
        """
        if level == MetricLevelEnum.MACRO:
            df = self.df_macro_level_metric()
        elif level == MetricLevelEnum.MICRO:
            df = self.df_micro_level_metric()
        elif level == MetricLevelEnum.WINDOW:
            df = self.df_window_level_metric()
        elif level == MetricLevelEnum.USER:
            df = self.df_user_level_metric()
        else:
            raise ValueError("Invalid level specified")

        if filter_algo:
            df = df.filter(like=filter_algo, axis=0)
        if filter_timestamp:
            df = df.filter(like=f"t={filter_timestamp}", axis=0)
        return df



import logging
from typing import Self

import numpy as np
from scipy.sparse import csr_matrix, hstack, vstack

from ...matrix import PredictionMatrix
from ..base import PopularityPaddingMixin, TopKAlgorithm


logger = logging.getLogger(__name__)


class MostPopular(TopKAlgorithm, PopularityPaddingMixin):
    """A popularity-based algorithm that considers all historical data."""

    IS_BASE: bool = False
    X_: csr_matrix | None = None  # Store all historical training data

    def _append_training_data(self, X: csr_matrix) -> None:
        """Append a new interaction matrix to the historical data.

        Args:
            X (csr_matrix): Interaction matrix to append
        """
        if self.X_ is None:
            raise ValueError("No existing training data to append to.")
        X_prev: csr_matrix = self.X_.copy()
        new_num_rows = max(X_prev.shape[0], X.shape[0])
        new_num_cols = max(X_prev.shape[1], X.shape[1])
        # Pad the previous matrix
        if X_prev.shape[0] < new_num_rows:  # Pad rows
            row_padding = csr_matrix((new_num_rows - X_prev.shape[0], X_prev.shape[1]))
            X_prev = vstack([X_prev, row_padding])
        if X_prev.shape[1] < new_num_cols:  # Pad columns
            col_padding = csr_matrix((X_prev.shape[0], new_num_cols - X_prev.shape[1]))
            X_prev = hstack([X_prev, col_padding])

        # Pad the current matrix
        if X.shape[0] < new_num_rows:  # Pad rows
            row_padding = csr_matrix((new_num_rows - X.shape[0], X.shape[1]))
            X = vstack([X, row_padding])
        if X.shape[1] < new_num_cols:  # Pad columns
            col_padding = csr_matrix((X.shape[0], new_num_cols - X.shape[1]))
            X = hstack([X, col_padding])

        # Merge data
        self.X_ = X_prev + X

    def _fit(self, X: csr_matrix) -> Self:
        if self.X_ is not None:
            self._append_training_data(X)
        else:
            self.X_ = X.copy()

        if not isinstance(self.X_, csr_matrix):
            raise ValueError("Training data is not initialized properly.")

        if self.X_.shape[1] < self.K:
            logger.warning("K is larger than the number of items.", UserWarning)

        self.sorted_scores_ = self.get_popularity_scores(self.X_)
        return self

    def _predict(self, X: PredictionMatrix) -> csr_matrix:
        intended_shape = (X.get_prediction_data().num_interactions, X.user_item_shape[1])

        # Vectorized: repeat the sorted scores for each prediction row
        data = np.tile(self.sorted_scores_, (intended_shape[0], 1))
        X_pred = csr_matrix(data)

        return X_pred



import logging

from recnexteval.matrix import InteractionMatrix
from .base import Setting
from .splitters import (
    NLastInteractionSplitter,
)


logger = logging.getLogger(__name__)


class LeaveNOutSetting(Setting):
    IS_BASE: bool = False

    def __init__(
        self,
        n_seq_data: int = 1,
        N: int = 1,
        seed: int = 42,
    ) -> None:
        super().__init__(seed=seed)
        self.n_seq_data = n_seq_data
        # we use top_K to denote the number of items to predict
        self.top_K = N

        logger.info("Splitting data")

        self._splitter = NLastInteractionSplitter(N, n_seq_data)

    def _split(self, data: InteractionMatrix) -> None:
        """Splits your dataset into a training, validation and test dataset
            based on the timestamp of the interaction.

        :param data: Interaction matrix to be split. Must contain timestamps.
        :type data: InteractionMatrix
        """

        self._background_data, future_interaction = self._splitter.split(data)
        # we need to copy the data to avoid modifying the background data
        past_interaction = self._background_data.copy()

        self._unlabeled_data, self._ground_truth_data = self.prediction_data_processor.process(
            past_interaction=past_interaction,
            future_interaction=future_interaction,
            top_K=self.top_K,
        )
        self._t_window = None



from warnings import warn

import numpy as np
from scipy.sparse import csr_matrix, lil_matrix

from ...matrix import InteractionMatrix
from ..base import Algorithm


class DecayPopularity(Algorithm):
    """A popularity-based algorithm with exponential decay over data from earlier time windows.
    """
    IS_BASE: bool = False

    def __init__(self, K: int = 200) -> None:
        super().__init__()
        self.K = K
        self.historical_data: list[csr_matrix] = []  # Store all historical training data
        self.num_items = 0  # Track the maximum number of items seen so far

    def _pad_matrix(self, matrix: csr_matrix, new_num_items: int) -> csr_matrix:
        """
        Pad a sparse matrix with zero columns to match the new number of items.

        :param matrix: The matrix to pad
        :type matrix: csr_matrix
        :param new_num_items: The target number of columns
        :type new_num_items: int
        :return: The padded matrix
        :rtype: csr_matrix
        """
        if matrix.shape[1] >= new_num_items:
            return matrix
        padding = csr_matrix((matrix.shape[0], new_num_items - matrix.shape[1]))
        return csr_matrix(np.hstack([matrix.toarray(), padding.toarray()]))

    def _expand_historical_data(self, new_num_items: int):
        """
        Expand all matrices in historical_data to match the new number of items.

        :param new_num_items: The updated number of items
        :type new_num_items: int
        """
        for i in range(len(self.historical_data)):
            if self.historical_data[i].shape[1] < new_num_items:
                self.historical_data[i] = self._pad_matrix(self.historical_data[i], new_num_items)

    def _fit(self, X: csr_matrix) -> "DecayPopularity":
        """
        Fit the model by applying decay to historical data and adding new data.

        :param X: Interaction matrix (users x items) for the current window
        :type X: csr_matrix
        """
        # Update the maximum number of items
        new_num_items = X.shape[1]
        if new_num_items > self.num_items:
            self._expand_historical_data(new_num_items)
            self.num_items = new_num_items

        # Append the new matrix (ensure it has the correct number of items)
        if X.shape[1] < self.num_items:
            X = self._pad_matrix(X, self.num_items)

        # Append new data to historical data
        self.historical_data.append(X)

        # Initialize decayed scores
        num_items = X.shape[1]
        if num_items < self.K:
            warn("K is larger than the number of items.", UserWarning)

        decayed_scores = np.zeros(num_items)

        # Apply decay to each historical matrix
        for i, matrix in enumerate(self.historical_data):
            # length 2, i = 0 -> 2-1-0 = 1, i = 1 -> 2-1-1 = 0
            # length 3, i = 0 -> 3-1-0 = 2, i = 1 -> 3-1-1 = 1, i = 2 -> 3-1-2 = 0
            decay_factor = np.exp(-(len(self.historical_data) - 1 - i))
            decayed_scores += matrix.sum(axis=0).A[0] * decay_factor

        normalized_scores = decayed_scores / decayed_scores.max()

        K = min(self.K, num_items)
        ind = np.argpartition(normalized_scores, -K)[-K:]
        a = np.zeros(num_items)
        a[ind] = normalized_scores[ind]
        self.decayed_scores_ = a
        return self

    def _predict(self, X: csr_matrix, predict_im: InteractionMatrix) -> csr_matrix:
        """
        Predict the K most popular item for each user scaled by the decay factor.
        """
        if predict_im is None:
            raise AttributeError("Predict frame with requested ID is required for Popularity algorithm")

        predict_frame = predict_im._df

        users = predict_frame["uid"].unique().tolist()
        known_item_id = X.shape[1]

        # predict_frame contains (user_id, -1) pairs
        max_user_id = predict_frame["uid"].max() + 1
        intended_shape = (max(max_user_id, X.shape[0]), known_item_id)

        X_pred = lil_matrix(intended_shape)
        X_pred[users] = self.decayed_scores_

        return X_pred.tocsr()



import logging

from .itemknn import ItemKNN


logger = logging.getLogger(__name__)


class ItemKNNRolling(ItemKNN):
    """Rolling version of ItemKNN algorithm.

    This class extends the ItemKNN algorithm to update the memory of the model
    to only keep the last window of interactions. The model is simply discarding
    all interactions that are older than the window size.
    """

    IS_BASE: bool = False



import pandas as pd

from .base import Dataset


class TestDataset(Dataset):
    """
    Test dataset.

    The test dataset is a dummy dataset that is used for testing purposes.
    """

    USER_IX = "user_id"
    """Name of the column in the DataFrame that contains user identifiers."""
    ITEM_IX = "item_id"
    """Name of the column in the DataFrame that contains item identifiers."""
    TIMESTAMP_IX = "timestamp"
    """Name of the column in the DataFrame that contains time of interaction in seconds since epoch."""
    RATING_IX = "rating"
    """Name of the column in the DataFrame that contains the rating a user gave to the item."""
    DEFAULT_FILENAME = "dummy_input.csv"

    def _download_dataset(self) -> None:
        pass

    def _load_dataframe(self) -> pd.DataFrame:
        """Load the raw dataset from file, and return it as a pandas DataFrame.

        Warning:
            This does not apply any preprocessing, and returns the raw dataset.

        Returns:
            The interaction data as a DataFrame with a row per interaction.
        """
        input_dict = {
            self.USER_IX: [1, 2, 3, 1, 2, 2, 4, 3, 3, 4, 5, 5, 5],
            self.ITEM_IX: [1, 1, 2, 3, 2, 3, 2, 1, 3, 3, 1, 2, 3],
            self.TIMESTAMP_IX: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10, 10],
            self.RATING_IX: [1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 1, 2, 3],
        }

        df = pd.DataFrame.from_dict(input_dict)
        return df

    def _fetch_dataset_metadata(self, user_id_mapping: pd.DataFrame, item_id_mapping: pd.DataFrame) -> None:
        pass



import logging
import logging.config
import os
import warnings
from collections.abc import Generator
from contextlib import contextmanager
from enum import Enum

import pyfiglet
import yaml

from recnexteval.utils.path import safe_dir
from recnexteval.utils.yaml_tool import create_config_yaml


class LogLevel(Enum):
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET

    @classmethod
    def from_string(cls, level: str) -> "LogLevel":
        """Return a LogLevel member from a case-insensitive string.

        Args:
            level: Name of the log level (case-insensitive).

        Returns:
            The corresponding :class:`LogLevel` enum member.
        """
        return cls[level.upper()]


def log_level(level: int | str | LogLevel) -> None:
    """Change the logging level for the root logger.

    Args:
        level: The logging level to set. May be a :class:`LogLevel` enum,
            a level name (str, case-insensitive), or an integer logging level.

    Returns:
        None
    """
    if isinstance(level, str):
        level = LogLevel.from_string(level)

    numeric_level = level.value if isinstance(level, LogLevel) else level

    logger = logging.getLogger()
    logger.setLevel(numeric_level)
    for handler in logger.handlers:
        handler.setLevel(numeric_level)


log_level_by_name = log_level  # Alias for convenience


def suppress_warnings() -> None:
    """Suppress all Python warnings.

    This will disable warning output by filtering all warnings and
    disabling logging's capture of warnings.

    Returns:
        None
    """
    logging.captureWarnings(False)
    warnings.filterwarnings("ignore")


def enable_warnings() -> None:
    """Enable Python warnings (reset to default behavior).

    This re-enables warning capture and resets any filters previously set.

    Returns:
        None
    """
    logging.captureWarnings(True)
    warnings.resetwarnings()


def suppress_specific_warnings(category: type[Warning]) -> None:
    """Suppress warnings of a specific category.

    Args:
        category: Warning class/type to suppress (for example, :class:`DeprecationWarning`).

    Returns:
        None
    """
    warnings.filterwarnings("ignore", category=category)


@contextmanager
def warnings_suppressed() -> Generator:
    """Context manager that temporarily suppresses all warnings.

    Yields:
        None: Warnings are suppressed inside the context block.
    """
    logging.captureWarnings(False)
    warnings.filterwarnings("ignore")
    try:
        yield
    finally:
        logging.captureWarnings(True)
        warnings.resetwarnings()


def prepare_logger(log_config_filename: str) -> dict:
    """Prepare and configure logging from a YAML file.

    This function locates or creates a logging configuration YAML file using
    :func:`recnexteval.utils.yaml_tool.create_config_yaml`, ensures the
    directory for the configured log file exists, writes an ASCII art header
    to the log file, and configures the Python logging system using
    :func:`logging.config.dictConfig`.

    Args:
        log_config_filename: Name of the logging configuration YAML file.

    Returns:
        dict: The parsed logging configuration dictionary.

    Raises:
        FileNotFoundError: If the resolved YAML configuration file cannot be found.
        ValueError: If there is an error parsing the YAML content.
    """
    _, yaml_file_path = create_config_yaml(log_config_filename)
    try:
        with open(yaml_file_path, "r") as stream:
            config = yaml.load(
                stream, Loader=yaml.FullLoader
            )
    except FileNotFoundError:
        raise FileNotFoundError(f"Configuration file not found at {yaml_file_path}.")
    except yaml.YAMLError as e:
        raise ValueError(f"Error parsing YAML configuration: {e}")

    # Get the log file path from the configuration
    log_file = config["handlers"]["file"]["filename"]

    # Ensure the log file directory exists
    dir_name = os.path.dirname(log_file)
    safe_dir(dir_name)

    # Write ASCII art to the log file
    with open(log_file, "w") as log:
        ascii_art = pyfiglet.figlet_format("recnexteval")
        log.write(ascii_art)
        log.write("\n")

    logging.config.dictConfig(config)
    logging.captureWarnings(True)
    return config



from .amazon import (
    AmazonBookDataset,
    AmazonMovieDataset,
    AmazonMusicDataset,
    AmazonSubscriptionBoxesDataset,
)
from .base import Dataset
from .lastfm import LastFMDataset
from .movielens import MovieLens100K
from .test import TestDataset
from .yelp import YelpDataset


__all__ = [
    "AmazonBookDataset",
    "AmazonMovieDataset",
    "AmazonMusicDataset",
    "AmazonSubscriptionBoxesDataset",
    "LastFMDataset",
    "MovieLens100K",
    "YelpDataset",
    "TestDataset",
    "Dataset",
]



"""Builder module for constructing evaluator objects.

This module provides builder classes for constructing evaluator objects in the
RecNextEval library. Builders follow the builder pattern to facilitate the
construction of evaluators with proper validation and error checking.

## Builder Overview

The builder pattern is used to construct complex evaluator objects step by step.
Builders ensure that all necessary components (settings, metrics, algorithms)
are properly configured before building the evaluator, preventing runtime errors.

## Available Builders

- `Builder`: Abstract base class for all builder implementations
- `EvaluatorPipelineBuilder`: Builder for pipeline evaluators that evaluate
  multiple algorithms on static data
- `EvaluatorStreamerBuilder`: Builder for streaming evaluators that evaluate
  algorithms on streaming data

## Using Builders

### Basic Pipeline Evaluation

To evaluate multiple algorithms on a static dataset using a pipeline evaluator:

```python
from recnexteval.evaluators.builder import EvaluatorPipelineBuilder
from recnexteval.settings import Setting
from recnexteval.datasets import AmazonMusicDataset

# Load dataset
dataset = AmazonMusicDataset()
data = dataset.load()

# Create setting
setting = Setting(data=data, top_K=10)
setting.split()

# Build evaluator
builder = EvaluatorPipelineBuilder(seed=42)
builder.add_setting(setting)
builder.set_metric_K(10)
builder.add_metric("PrecisionK")
builder.add_metric("RecallK")
builder.add_algorithm("MostPopular")
builder.add_algorithm("RecentPop", params={"K": 10})

evaluator = builder.build()
results = evaluator.evaluate()
```

### Streaming Evaluation

To evaluate algorithms on streaming data:

```python
from recnexteval.evaluators.builder import EvaluatorStreamerBuilder
from recnexteval.settings import StreamingSetting
from recnexteval.datasets import AmazonMusicDataset

# Load dataset
dataset = AmazonMusicDataset()
data = dataset.load()

# Create streaming setting
setting = StreamingSetting(data=data, top_K=10, window_size=1000)
setting.split()

# Build streaming evaluator
builder = EvaluatorStreamerBuilder(seed=42)
builder.add_setting(setting)
builder.set_metric_K(10)
builder.add_metric("HitK")
builder.add_metric("NDCGK")

evaluator = builder.build()
# The evaluator can now process streaming data
```

### Advanced Configuration

Builders support advanced configuration options:

```python
from recnexteval.evaluators.builder import EvaluatorPipelineBuilder

builder = EvaluatorPipelineBuilder(
    ignore_unknown_user=False,  # Don't ignore unknown users
    ignore_unknown_item=True,   # Ignore unknown items
    seed=123
)

builder.add_setting(setting)
builder.set_metric_K(20)

# Add multiple metrics
metrics = ["PrecisionK", "RecallK", "DCGK", "NDCGK", "HitK"]
for metric in metrics:
    builder.add_metric(metric)

# Add algorithms with custom parameters
builder.add_algorithm("ItemKNN", params={"K": 50, "similarity": "cosine"})
builder.add_algorithm("DecayPop", params={"decay_factor": 0.9})

evaluator = builder.build()
```

## Extending the Framework

To create custom builders, inherit from the `Builder` base class and implement
the `build()` method. Ensure to call `super().__init__()` and implement proper
validation in `_check_ready()`.
"""

from .base import Builder
from .pipeline import EvaluatorPipelineBuilder
from .stream import EvaluatorStreamerBuilder


__all__ = [
    "Builder",
    "EvaluatorPipelineBuilder",
    "EvaluatorStreamerBuilder",
]
